__all__ = ["FileFCF"]

from .. import DataFile
from collections import OrderedDict
import re
import a99

class FileFCF(DataFile):
    """
    File containing Franck-Condon Factors (FCFs)

    Usage: attribute "fcfs" is a dictionary accessed by key (vl, v2l)

    **History:**

    This file is the output of the Fortran code that calculated FCFs for Bruno Castilho's thesis.

    In 2015, I modified this Fortran code to compile with gfortran on request from
    Prof. Amaury Augusto de Almeida, but I believe that the code remains unpublished.

    Nevertheless, Bruno left several output files in his directory ATMOS:/wrk4/bruno/Mole/Fc
    containing tabulated FCFs for several molecules. The reason why I created this class is to read
    these files.
    """

    attrs = ["fcfs"]

    @property
    def num_lines(self):
        return len(self)

    def __len__(self):
        return len(self.fcfs)

    def __getitem__(self, item):
        return self.fcfs[item]

    def __init__(self):
        DataFile.__init__(self)

        self.fcfs = OrderedDict()


    def _do_load(self, filename):
        # File part of interest looks like this:

        #                                 V"   V""  FRANCK CONDON FACTOR     R-CENTROID
        #
        #                                  0   0    .8809749E+00             .1159111E+01
        #                                  0   1    .6597226E-01             .1573085E+01
        #                                  0   2    .4967269E-01             .1309538E+01
        #                                  0   3    .4893512E-06             .5487973E+02
        #                                  0   4    .1967671E-02             .1092197E+01
        #                                  0   5    .5098846E-03             .7344212E+00
        #                                  0   6    .3264353E-03             .8154120E+00
        #                                  0   7    .1959470E-03             .8020893E+00
        #                                  0   8    .1221533E-03             .7933001E+00
        #                                  0   9    .7849922E-04             .7871405E+00
        #                                  0  10    .5130633E-04             .7806426E+00
        #
        #                                 V"   V""  FRANCK CONDON FACTOR     R-CENTROID
        #
        #                                  1   0    .7950073E-01             .9007177E+00
        #                                  1   1    .7301335E+00             .1176593E+01
        #                                  1   2    .5329854E-01             .1873632E+01
        #                                  1   3    .1172008E+00             .1293195E+01
        #                                  1   4    .9687749E-03            -.1106724E+01
        #                                  1   5    .8699398E-02             .1027710E+01
        #                                  1   6    .3409328E-02             .7787252E+00
        #                                  1   7    .2211213E-02             .8098937E+00
        #                                  1   8    .1418452E-02             .8016906E+00
        #                                  1   9    .9225228E-03             .7913423E+00
        #                                  1  10    .6110833E-03             .7833238E+00
        # 1
        #
        #                                 V"   V""  FRANCK CONDON FACTOR     R-CENTROID
        #
        #                                  2   0    .1939492E-01             .1045730E+01
        #                                  2   1    .4528844E-01             .1349549E+01
        #                                  2   2    .1031371E+00             .1003365E+01
        #                                  2   3    .4823686E+00             .1247711E+01
        #                                  2   4    .7587573E-01             .2082369E+01
        #                                  2   5    .2123286E+00             .1379030E+01
        #                                  2   6    .2887595E-02            -.1045935E+01
        #                                  2   7    .1985530E-01             .9975332E+00
        #                                  2   8    .1103128E-01             .8153041E+00
        #                                  2   9    .7418684E-02             .8001242E+00
        #                                  2  10    .5176848E-02             .7976580E+00
        #

        # magic characters
        MAGIC = 'V"   V""  FRANCK CONDON FACTOR     R-CENTROID'

        # possible states for parsing file
        EXP_MAGIC = 0 # Expecting magic characters
        EXP_BLANK = 1 # Expecting blank line
        EXP_DATA = 2 # Expecting data, "1*" or blank line

        # regular expression for extraction the data
        rec = re.compile("\s*(\d+)\s*(\d+)\s*([\deE\.\+\-]+)")


        with open(filename, "r") as h:
            found = False
            state = EXP_MAGIC
            i = 0
            try:
                for line in h:
                    i += 1
                    if state == EXP_MAGIC:
                       if line.strip() == MAGIC:
                           state = EXP_BLANK
                           found = True

                    elif state == EXP_BLANK:
                        if line.strip() == "":
                            state = EXP_DATA
                    elif state == EXP_DATA:
                        if line.strip() == "" or line[0] == "1":
                            state = EXP_MAGIC
                        else:
                            m = rec.match(line)
                            if m is None:
                                raise RuntimeError("Could not extract (vl, v2l, fcf)")
                            vl, v2l, fcf = m.groups()
                            vl = int(vl)
                            v2l = int(v2l)

                            if (vl, v2l) in self.fcfs:
                                a99.get_python_logger().warning("Repeated (vl, v2l): ({}, {})".format(vl, v2l))

                            fcf = float(fcf)
                            self.fcfs[(vl, v2l)] = fcf
            except Exception as e:
                raise RuntimeError("Error in line {}".format(i)) from e


            if not found:
                raise RuntimeError("File does not appear to be a FileFCF")




__all__ = ["SB_Rubberband", "SB_AddNoise", "SB_FNuToFLambda", "SB_FLambdaToFNu", "SB_ElementWise",
           "SB_Extend", "SB_Cut", "SB_Normalize", "SB_ConvertYUnit", "SB_Add", "SB_Mul",
           "SB_MulByLambda", "SB_DivByLambda"]


import numpy as np
import a99
import copy
from . import SpectrumBlock
import astropy.units as u
from ... import physics as ph
from .. import basic

class SB_Rubberband(SpectrumBlock):
    """
    Convex polygonal line (aka "rubberband")

    Args:
        flag_upper=True: whether to stretch rubberband from above the
            spectrum; otherwise, stretches line from below

    Stretches a polygonal line from below/above the spectrum. The vertices of this multi-segment
    line will touch "troughs" of the spectrum without crossing the spectrum

    This was inspired on -- but is not equivalent to -- OPUS Rubberband baseline correction [1].
    However, this one is parameterless, whereas OPUS RBBC asks for a number of points

    References:
        [1] Bruker Optik GmbH, OPUS 5 Reference Manual. Ettlingen: Bruker, 2004.
    """

    def __init__(self, flag_upper=True):
        SpectrumBlock.__init__(self)
        # Upper or lower rubberband
        self.flag_upper = flag_upper

    def _do_use(self, inp):
        output = self._new_output()
        y = inp.y
        if self.flag_upper:
            y = -y
        output.y = basic.rubberband(y)
        if self.flag_upper:
            output.y = -output.y
        return output


class SB_AddNoise(SpectrumBlock):
    """
    Adds normally distributed (Gaussian) random noise

    Args:
        std=1.: standard deviation of Gaussian noise
    """
    def __init__(self, std=1.):
        SpectrumBlock.__init__(self)
        # Standard deviation of noise
        self.std = std

    def _do_use(self, inp):
        n = len(inp)
        output = self._new_output()
        output.y = np.copy(inp.y) + np.random.normal(0, self.std, n)
        return output


class SB_ElementWise(SpectrumBlock):
    """
    Applies specified function to spectrum flux

    Args:
        func: a function that takes a vector (_i.e._, NumPy 1-D array) as input. It must return
            vector of same dimension as input.NumPy ufunc's are suited for this.
            Examples:
                np.square
                np.exp
            It also be a lambda using list comprehension, for example:
                lambda v: [x**2 for x in v]
    """

    def __init__(self, func):
        SpectrumBlock.__init__(self)
        self.func = func

    def _do_use(self, inp):
        output = self._new_output()
        output.wavelength = np.copy(inp.wavelength)
        output.flux = self.func(inp.flux)
        if len(output.flux) != len(output.wavelength):
            raise RuntimeError(
                "func returned vector of length %d, but should be %d" % (len(output.flux), len(output.wavelength)))
        if not isinstance(output.flux, np.ndarray):
            output.flux = np.array(output.flux)
        return output


class SB_Extend(SpectrumBlock):
    """
    Extends to left and/or right side

    Args:
        fraction: amount relative to number of points. Note that this
                    applies individually to left and right (see below)
        flag_left: whether to extend by fraction to left
        flag_right: whether to extend by fraction to right
        fill_mode='poly_baseline': how to fill the new points. Valid values:
            'poly_baseline': The y-value (left/right) to use is found by using a "coarse"
                                2nd-order polynomial baseline. The baseline is "coarse" because it
                                does not allow for many iterations until the baseline is found
            'zero': Fills with zero

    Examples:
        SB_Extend(.1, True, True)  # if original has 100 points, resulting will have 120 points

        SB_Extend(.1, True, False)  # if original has 100 points, resulting will have 110 points
    """

    def __init__(self, fraction=.1, flag_left=True, flag_right=False, fill_mode='poly_baseline'):
        SpectrumBlock.__init__(self)
        self.fraction = fraction
        self.flag_left = flag_left
        self.flag_right = flag_right
        if not fill_mode in ('poly_baseline', 'zero'):
            raise RuntimeError("Invalid fill mode: '%s'" % fill_mode)
        self.fill_mode = fill_mode

    def _do_use(self, inp):
        from f311 import explorer as ex

        # TODO work with delta_lambda not constant
        output = self._copy_input(inp)

        if not (self.flag_left or self.flag_right):
            return output
        num_add = int(self.fraction*len(output.wavelength))
        if num_add <= 0:
            return output

        x_left, x_right, y_left, y_right = np.array([]), np.array([]), np.array([]), np.array([])

        if self.fill_mode == 'poly_baseline':
            baseline = -ex.poly_baseline(-output.y, 2, maxit=15)
            fill_l = baseline[0]
            fill_r = baseline[-1]
        elif self.fill_mode == 'zero':
            fill_l = 0
            fill_r = 0
        else:
            raise RuntimeError("Invalid fill mode: '%s'" % self.fill_mode)

        if self.flag_left:
            x_left = np.arange(num_add)*output.delta_lambda+(output.x[0]-output.delta_lambda*num_add)
            y_left = np.ones(num_add)*fill_l

        if self.flag_right:
            x_right = np.arange(num_add) * output.delta_lambda + (output.x[-1] + output.delta_lambda)
            y_right = np.ones(num_add)*fill_r

        output.x = np.concatenate((x_left, output.x, x_right))
        output.y = np.concatenate((y_left, output.y, y_right))

        return output


class SB_Cut(SpectrumBlock):
    """
    Cuts spectrum given a wavelength interval

    Args:
        l0: initial wavelength
        lf: final wavelength
    """

    def __init__(self, l0, lf):
        SpectrumBlock.__init__(self)
        if l0 >= lf:
            raise ValueError("l0 must be lower than lf")
        self.l0 = l0
        self.lf = lf

    def __repr__(self):
        return "SB_Cut({}, {})".format(self.l0, self.lf)

    def _do_use(self, inp):
        idx0 = np.argmin(np.abs(inp.x - self.l0))
        idx1 = np.argmin(np.abs(inp.x - self.lf))
        out = self._copy_input(inp)
        out.x = out.x[idx0:idx1]
        out.y = out.y[idx0:idx1]
        return out


class SB_Normalize(SpectrumBlock):
    """
    Normalizes spectrum according to specified method

    Args:
        method --
            "01": normalizes between 0 and 1
            "1": scales to have maximum at 1, keeping the 0 where it is
    """

    def __init__(self, method="01"):
        SpectrumBlock.__init__(self)
        choices = ["01", "1"]
        if not method in choices:
            raise ValueError("Invalid normalization method; choices: %s" % str(choices))
        self.method = method

    def __repr__(self):
        return "SB_Normalize('{}')".format(self.method)

    def _do_use(self, inp):
        out = self._copy_input(inp)
        if self.method == "01":
            miny, maxy = np.min(out.y), np.max(out.y)
            if miny == maxy:
                raise RuntimeError("Cannot normalize, y-span is ZERO")
            out.y = (out.y - miny) / (maxy - miny)

        elif self.method == "1":
            maxy = np.max(out.y)
            out.y = out.y / maxy
        return out


# # Temporarily suspended (changing x-axis has implications in SpectrumList.wavelength vector
#
# class SB_ConvertXUnit(SpectrumBlock):
#     """
#     Converts x-axis unit
#
#     Args:
#         new_unit: astropy.units.unit or str
#     """
#
#     def __init__(self, new_unit):
#         SpectrumBlock.__init__(self)
#         self.unit = u.Unit(new_unit)
#
#     def __repr__(self):
#         return "{}({})".format(self.__class__.__name__, repr(self.unit))
#
#     def _do_use(self, inp):
#         qty = inp.x * inp.xunit
#         out = self._copy_input(inp)
#         out.x = qty.to(self.unit).value
#         out.unit = copy.deepcopy(self.unit)
#         return out


class SB_ConvertYUnit(SpectrumBlock):
    """
    Converts y-axis unit

    Args:
        new_unit: astropy.units.unit or str
    """

    def __init__(self, new_unit):
        SpectrumBlock.__init__(self)
        self.unit = u.Unit(new_unit)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, repr(self.unit))

    def _do_use(self, inp):
        # qty = inp.y * inp.yunit
        out = self._copy_input(inp)
        out.y = out.yunit.to(self.unit, out.y, equivalencies=u.spectral_density(out.xunit, out.x))
        out.yunit = copy.deepcopy(self.unit)
        return out


class SB_FNuToFLambda(SB_ConvertYUnit):
    """
    Flux-nu to flux-lambda conversion. Assumes the wavelength axis is in angstrom

    Formula:
        f_lambda = f_nu*(nu/lambda) = f_nu*c/lambda**2

        where
            lambda is the wavelength in cm,
            c is the speed of light in cm/s
            f_lambda has irrelevant unit for this purpose
    """

    def __init__(self):
        SB_ConvertYUnit.__init__(self, ph.flambda)


class SB_FLambdaToFNu(SB_ConvertYUnit):
    """
    Flux-nu to flux-lambda conversion. Assumes the wavelength axis is in angstrom

    Formula:
        f_nu = f_lambda*(lambda/nu) = f_lambda*lambda**2/c
    """
    def __init__(self):
        SB_ConvertYUnit.__init__(self, ph.fnu)



class SB_Mul(SpectrumBlock):
    """
    Multiplies y-values by constant value

    Args:
        value
    """

    def __init__(self, value=1.):
        SpectrumBlock.__init__(self)
        self.value = value

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, repr(self.value))

    def _do_use(self, inp):
        # qty = inp.y * inp.yunit
        out = self._copy_input(inp)
        out.y *= self.value
        return out


class SB_Add(SpectrumBlock):
    """
    Adds constant value to y-values

    Args:
        value
    """

    def __init__(self, value=0.):
        SpectrumBlock.__init__(self)
        self.value = value

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, repr(self.value))

    def _do_use(self, inp):
        out = self._copy_input(inp)
        out.y += self.value
        return out


class SB_MulByLambda(SpectrumBlock):
    """Multiplies each y-point by its wavelength"""

    def _do_use(self, inp):
        out = self._copy_input(inp)
        out.y *= out.x
        return out


class SB_DivByLambda(SpectrumBlock):
    """Divides each y-point by its wavelength"""

    def _do_use(self, inp):
        out = self._copy_input(inp)
        out.y /= out.x
        return out




__all__ = ["XFileSpectrumList"]


import matplotlib.pyplot as plt
import os
import os.path
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from .a_WFileSpectrumList import *
import a99
import f311.filetypes as ft
from ..a_XFileMainWindow import *


class XFileSpectrumList(XFileMainWindow):
    def __init__(self, parent=None, fileobj=None):
        XFileMainWindow.__init__(self, parent)

        def keep_ref(obj):
            self._refs.append(obj)
            return obj

        self.setWindowTitle(a99.get_window_title("Spectrum List Editor"))

        # # Synchronized sequences
        _VVV = ft.FileSpectrumList.description
        self.text_tab[0] =  "FileSpectrumList editor (Alt+&1)"
        self.tabWidget.setTabText(0, self.text_tab[0])
        text_save_as="Save %s as..." % _VVV
        text_load="Load %s" % _VVV
        cls=ft.FileSpectrumList
        clssft.classes_collection()  # file types that can be opened
        clsstuple([ft.FileSpectrumList, ft.FileFullCube]+ft.classes_sp())  # file types that can be opened
        wild="*.splist"

        lv = keep_ref(QVBoxLayout(self.gotting))
        ce = self.ce = WFileSpectrumList(self)
        lv.addWidget(ce)
        ce.changed.connect(self.on_tab0_file_edited)
        self.editors[0] = ce

        # # Adds spectrum collection actions to menu
        self.menuBar().addMenu(self.ce.menu_actions)

        # # Loads default file by default
        if os.path.isfile(ft.FileSpectrumList.default_filename):
            f = ft.FileSpectrumList()
            f.load()
            self.ce.load(f)

        if fileobj is not None:
            self.load(fileobj)


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Qt override

    def closeEvent(self, event):
        flag_exit, ff = True, []
        for ed, flag_changed in zip(self.editors, self.flags_changed):
            if ed and ed.f and flag_changed:
                ff.append(ed.f.description)

        if len(ff) > 0:
            s = "Unsaved changes\n  -"+("\n  -".join(ff))+"\n\nAre you sure you want to exit?"
            flag_exit = a99.are_you_sure(True, event, self, "Unsaved changes", s)
        if flag_exit:
            plt.close("all")

    def keyPressEvent(self, evt):
        incr = 0
        if evt.modifiers() == Qt.ControlModifier:
            n = self.tabWidget.count()
            if evt.key() in [Qt.Key_PageUp, Qt.Key_Backtab]:
                incr = -1
            elif evt.key() in [Qt.Key_PageDown, Qt.Key_Tab]:
                incr = 1
            if incr != 0:
                new_index = self._get_tab_index() + incr
                if new_index < 0:
                    new_index = n-1
                elif new_index >= n:
                    new_index = 0
                self.tabWidget.setCurrentIndex(new_index)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Slots for Qt library signals

    # def on_show_rm(self):
    #     if self._manager_form:
    #         self._manager_form.show()
    #         self._manager_form.raise_()
    #         self._manager_form.activateWindow()

    def on_tab0_file_edited(self):
        self._on_me_changed()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Protected methods to be overriden or used by descendant classes

    def _on_me_changed(self):
        index = self._get_tab_index()
        self.flags_changed[index] = True
        self._update_tab_texts()






__all__ = ["SLB_UseSpectrumBlock", "SLB_ExtractContinua"]

import numpy as np
import a99
import copy
from .basic import *
from . import sb

class SLB_UseSpectrumBlock(SpectrumListBlock):
    """Calls sblock.use() for each individual spectrum"""

    def __init__(self, sblock=None):
        SpectrumListBlock.__init__(self)
        self.sblock = sblock

    def _do_use(self, inp):
        output = self._new_output()
        for i, sp in enumerate(inp.spectra):
            output.add_spectrum(self.sblock.use(sp))
        return output


class SLB_ExtractContinua(SpectrumListBlock):
    """Calculates upper envelopes and subtracts mean(noise std)"""

    # TODO this is not a great system. Just de-noising could substantially improve the extracted continua

    def _do_use(self, inp):
        from . import gb
        output = SLB_UseSpectrumBlock(sb.SB_Rubberband(flag_upper=True)).use(inp)
        spectrum_std = gb.GB_UseNumPyFunc(func=np.std).use(inp)
        mean_std = np.mean(spectrum_std.spectra[0].y)
        for spectrum in output.spectra:
            spectrum.y -= mean_std * 3
        return output


class SLB_UseGroupBlock(SpectrumListBlock):
    """
    "Group by" operation using a GroupBlock

    Args:
        block: a GroupBlock
        group_by: sequence of spectrum "more_headers" fieldnames.
                  If not passed, will treat the whole SpectrumCollection as a single gb.py.
                  If passed, will split the collection in groups and perform the "merge down" operations separately
                  for each gb.py

    Returns: (SpectrumCollection containing query result, list of error strings)
    """

    def __init__(self, block, group_by=None):
        SpectrumListBlock.__init__(self)
        assert isinstance(block, GroupBlock)
        self.block = block
        self.group_by = group_by

    def _do_use(self, inp):

        # Creates the groups
        if not self.group_by:
            out = self.block.use(inp)
        else:
            groups = []

            grouping_keys = [
                tuple([spectrum.more_headers.get(fieldname) for fieldname in self.group_by]) for
                spectrum in inp.spectra]

            unique_keys = list(set(grouping_keys))
            unique_keys.sort()
            sk = list(zip(inp.spectra, grouping_keys))
            for unique_key in unique_keys:
                group = ft.SpectrumList()
                for spectrum, grouping_key in sk:
                    if grouping_key == unique_key:
                        group.add_spectrum(spectrum)
                groups.append(group)

            out = self._new_output()
            out.fieldnames = self.group_by  # new SpectrumList will have the gb.py field names

            # Uses block in each gb.py
            for group in groups:
                splist = self.block.use(group)

                # copies "gb.py by" fields from first input spectrum to output spectrum
                sp = splist.spectra[0]
                for fieldname in self.group_by:
                    sp.more_headers[fieldname] = group.spectra[0].more_headers[fieldname]

                out.merge_with(splist)

        return out



#!/usr/bin/env python


"""
Draws chart showing spectral lines of interest, spectrograph wavelength ranges, ESO atmospheric model, etc.

Two modes are available:
  - GUI mode (default): opens a GUI allowing for setup parameters
  - Plot mode (--plot): plots the chart directly in default way
"""


import sys
import argparse
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator, FormatStrFormatter
import matplotlib as mpl
import numpy as np
import numbers
from PyQt5.QtWidgets import *
import logging
from f311 import aosss as ao
from f311 import physics as ph
import a99


a99.logging_level = logging.INFO
a99.flag_log_file = True



###############################################################################
# # External setup
mpl.rcParams['axes.linewidth'] = 2


###############################################################################
# # local class & function library
class MyLine(object):
    def __init__(self, name, wl, width=0, position=0, reference=None):
        self.name = name
        self.wl = wl if isinstance(wl, (list, tuple)) else [wl]
        self.width = width
        self.reference = reference  # paper etc
        if width > 0 and not isinstance(wl, numbers.Number):
            raise RuntimeError("Either specify list of lines or width")
        self.position = position

def lambda_shift(lambda_0, v_radial):
    """Calculates delta_lambda due to Doppler effect given emitter wavelength and radial velocity in km/s.
    Positive radial velocity means running away from us."""
    C = 300000  # light speed in km/s
    return lambda_0*(C+v_radial)/C

class MyCoverage(object):
    """Stores data for the wavelength coverage of a certain instrument/mode etc"""
    def __init__(self, name, l0lf):
        self.name = name
        self.l0lf = l0lf  # sequence of tuples

###############################################################################
# # Data definition
NUM_SLOTS = 10
LAST_BAND = "K"
COLOR_COVERAGE = np.array([.5, .5, .5])
COLOR_GRID = np.array((.5, .5, .5))
COLOR_BAND = [0., .3, 0.]
COLOR_SPECTRUM = [.2, .2, .2]
COLOR_TELLURIC_BOX = [182./255, 234./255, 232./255]
COLOR_TEXT = (.25, .2, .2)

HA_LAMBDA = 6562.8
HA_WIDTH = lambda_shift(HA_LAMBDA, 400) - lambda_shift(HA_LAMBDA, -400)

ll = [MyLine("CaHK", 3900),
      MyLine("Sr", 4050),
      MyLine("C", 4300),
      MyLine("Mgb", [5167.32, 5172.68, 5183.6]),
      MyLine("NaD", [5889.95, 5895.92]),
      MyLine("CaI", 6162),
      MyLine(r"H$\alpha$", 6562.8, width=HA_WIDTH),
      MyLine("OI triplet", [7777.194, 7777.417, 7777.539]),
      MyLine("CaII triplet", [8498, 8542, 8662]),
      MyLine("He II", [1640]),
      MyLine("Lyman limit", 912.),
      MyLine("O II", 3727, reference="MPuech 2008"),
      MyLine("OH", 3064, reference="Castilho (Thesis) 1999"),
      MyLine("Be", 3131, reference="Castilho (Thesis) 1999"),
      MyLine("NH", 3360, reference="Castilho (Thesis) 1999"),
      MyLine("FeI, TiI, SiI, MgI", [11896, 11953, 11988, 11995, 12035, 12107, 11831, 12087, 11597, 11611, 11641, 11693, 11783, 11887, 11976], reference="Cevans 2011"),  # apud Davies et. al (2010)
      # MyLine("FeI", 11887, reference="Davies et. al (2010)"),
      # MyLine("FeI", 11641, reference="Davies et. al (2010)"),
      # MyLine("FeI", 11611, reference="Davies et. al (2010)"),
      # MyLine("MgI", 12087, reference="Davies et. al (2010)"),
      # MyLine("MgI", 11831, reference="Davies et. al (2010)"),
      # MyLine("MgI", 11831, reference="Davies et. al (2010)"),
      ]

cc = [MyCoverage("HMM", [(4000, 18000)]),
      MyCoverage("HDM (essential)", [(10000, 18000)]),
      MyCoverage("HDM (desirable)", [(8000, 25000)]),
      MyCoverage("IGM (essential)", [(4000, 8000)]),
      MyCoverage("IGM (desirable)", [(3700, 10000)])
      ]


# ###############################################################################
# def draw(fig, telluric_spectra, redshift=0):
#     l0, lf = 3000, a99.get_ubv_bandpass("K").lf
#     x =  np.logspace(np.log10(l0), np.log10(lf), 1000, base=10.)
#
#     ax = fig.gca()
#     ax.set_axisbelow(True)
#
#     # # grid
#     #   ====
#     # http://matplotlib.org/examples/pylab_examples/major_minor_demo1.html
#     majorLocator_x = MultipleLocator(1000)
#     majorFormatter_x = FormatStrFormatter('%d')
#     minorLocator_x = MultipleLocator(250)
#     ax.xaxis.set_major_locator(majorLocator_x)
#     ax.xaxis.set_major_formatter(majorFormatter_x)
#     # for the minor ticks, use no labels; default NullFormatter
#     ax.xaxis.set_minor_locator(minorLocator_x)
#
#     majorLocator_y = MultipleLocator(1.)
#     majorFormatter_y = FormatStrFormatter('')
#     # minorLocator_x = MultipleLocator(250)
#     ax.yaxis.set_major_locator(majorLocator_y)
#     ax.yaxis.set_major_formatter(majorFormatter_y)
#     # for the minor ticks, use no labels; default NullFormatter
#     # ax.xaxis.set_minor_locator(minorLocator_x)
#
#     g0 = plt.grid(True, which='major', axis='x', linewidth=2, linestyle=':', color=COLOR_GRID)
#     plt.grid(True, which='major', axis='y', linewidth=1, linestyle='-', color=COLOR_GRID*1.5)
#     plt.grid(True, which='minor', axis='x', linestyle=':', color=COLOR_GRID)
#
#
#     # # bands
#     #   =====
#     for band_name, bandpass in ph.get_ubv_bandpasses_dict().items():
#         y = bandpass.ufunc()(x)*.75
#         plt.plot(x, y, label=band_name, c=COLOR_BAND)
#         idx_max = np.argmax(y)
#         ax.annotate(band_name, xy=(x[idx_max], y[idx_max]+.1),
#                     horizontalalignment="center", verticalalignment="center",
#                     color=COLOR_TEXT)
#
#         if band_name == LAST_BAND:
#             break
#
#     # # Telluric features
#     Y_TOP = 2
#     Y_BOTTOM = Y_TOP-1+0.01
#     SPECTRUM_HEIGHT = 0.9  # Height for the spectrum to span in the chart
#     ax.annotate("Telluric features", xy=[l0, Y_TOP-.02], color=COLOR_TEXT, verticalalignment="top")
#     for sp in telluric_spectra:
#         # let ymin=0 (maybe there is no zero transmission in the spectrum,
#         # it will be good to see that frequencies are not completely attenuated)
#         ymax, ymin = max(sp.y), 0
#         x0, x1, y0, y1 = sp.x[0], sp.x[-1], Y_BOTTOM, Y_BOTTOM+SPECTRUM_HEIGHT
#         ax.fill_between([x0, x1], [y1, y1], [y0, y0], color=COLOR_TELLURIC_BOX)
#         ax.plot(sp.x, (sp.y-ymin)*SPECTRUM_HEIGHT+Y_BOTTOM, c=COLOR_SPECTRUM)
#
#
#     # # Chemical lines of interest (redshifted or not)
#     Y_TOP = NUM_SLOTS
#     COLOR_LINES = np.array([.3, 0., .3])
#     for line in ll:
#         wl = [x*(1+redshift) for x in line.wl]
#         if line.width > 0:
#             width = line.width*(1+redshift)
#             plt.fill_between([wl[0]-width/2, wl[0]+width/2],
#                              [Y_TOP, Y_TOP], [0, 0], color=COLOR_LINES, alpha=1)
#             x_ann = wl[0]+width/2
#         else:
#             for x in wl:
#                 plt.plot([x, x], [0, Y_TOP], c=COLOR_LINES, alpha=1)
#             x_ann = max(wl)
#         plt.annotate(line.name, xy=(x_ann, Y_TOP-.2), rotation=-90, color=COLOR_TEXT)
#
#
#     # # coverages
#     #   =========
#     HEIGHT = .7
#     y = NUM_SLOTS-(1-HEIGHT)/2-1
#     for coverage in cc:
#         for interval in coverage.l0lf:
#             x0, x1 = interval
#             plt.plot([x0, x1, x1, x0, x0], [y, y, y-HEIGHT, y-HEIGHT, y], color=COLOR_COVERAGE*.5, alpha=.8)
#             plt.fill_between(interval, [y, y], [y-HEIGHT, y-HEIGHT], color=COLOR_COVERAGE, alpha=.6, hatch='//')
#             x_ann = interval[1]
#         plt.annotate(coverage.name, xy=(x_ann+40, y-HEIGHT/2), color=COLOR_TEXT, verticalalignment="center")
#         y -= 1
#
#     plt.xlim([l0, lf])
#     plt.ylim([0, NUM_SLOTS+.01])
#     plt.xlabel("Wavelength ($\AA$)")
#     plt.tight_layout()
#


###############################################################################
def draw(fig, sp_sky, redshift=0):
    l0, lf = 3000, ph.get_ubv_bandpass("K").lf
    x =  np.logspace(np.log10(l0), np.log10(lf), 1000, base=10.)

    ax = fig.gca()
    ax.set_axisbelow(True)

    # # grid
    #   ====
    # http://matplotlib.org/examples/pylab_examples/major_minor_demo1.html
    majorLocator_x = MultipleLocator(1000)
    majorFormatter_x = FormatStrFormatter('%d')
    minorLocator_x = MultipleLocator(250)
    ax.xaxis.set_major_locator(majorLocator_x)
    ax.xaxis.set_major_formatter(majorFormatter_x)
    # for the minor ticks, use no labels; default NullFormatter
    ax.xaxis.set_minor_locator(minorLocator_x)

    majorLocator_y = MultipleLocator(1.)
    majorFormatter_y = FormatStrFormatter('')
    # minorLocator_x = MultipleLocator(250)
    ax.yaxis.set_major_locator(majorLocator_y)
    ax.yaxis.set_major_formatter(majorFormatter_y)
    # for the minor ticks, use no labels; default NullFormatter
    # ax.xaxis.set_minor_locator(minorLocator_x)

    g0 = plt.grid(True, which='major', axis='x', linewidth=2, linestyle=':', color=COLOR_GRID)
    plt.grid(True, which='major', axis='y', linewidth=1, linestyle='-', color=COLOR_GRID*1.5)
    plt.grid(True, which='minor', axis='x', linestyle=':', color=COLOR_GRID)


    # # bands
    #   =====
    for band_name, bandpass in ph.get_ubv_bandpasses_dict().items():
        y = bandpass.ufunc()(x)*.75
        plt.plot(x, y, label=band_name, c=COLOR_BAND)
        idx_max = np.argmax(y)
        ax.annotate(band_name, xy=(x[idx_max], y[idx_max]+.1),
                    horizontalalignment="center", verticalalignment="center",
                    color=COLOR_TEXT)

        if band_name == LAST_BAND:
            break

    # # ESO sky model
    if sp_sky:
        tt = ["Atmospheric emission", "Atmospheric transmission"]
        y_top = 3
        SPECTRUM_HEIGHT = 0.9  # Height for the spectrum to span in the chart
        for i, sp in enumerate(sp_sky):
            y_bottom = y_top-1+0.01
            # let ymin=0 (maybe there is no zero transmission in the spectrum,
            ymax, ymin = max(sp.y), 0
            y = sp.y/ymax
            x0, x1, y0, y1 = sp.x[0], sp.x[-1], y_bottom, y_bottom+SPECTRUM_HEIGHT
            ax.annotate(tt[i], xy=[l0, y_top-.42], color=COLOR_TEXT, verticalalignment="top")
            ax.fill_between([x0, x1], [y1, y1], [y0, y0], color=COLOR_TELLURIC_BOX)
            ax.plot(sp.x, (y-ymin)*SPECTRUM_HEIGHT+y_bottom, c=COLOR_SPECTRUM)

            y_top -= 1


    # # Chemical lines of interest (redshifted or not)
    Y_TOP = NUM_SLOTS
    COLOR_LINES = np.array([.3, 0., .3])
    for line in ll:
        wl = [x*(1+redshift) for x in line.wl]
        if line.width > 0:
            width = line.width*(1+redshift)
            plt.fill_between([wl[0]-width/2, wl[0]+width/2],
                             [Y_TOP, Y_TOP], [0, 0], color=COLOR_LINES, alpha=1)
            x_ann = wl[0]+width/2
        else:
            for x in wl:
                plt.plot([x, x], [0, Y_TOP], c=COLOR_LINES, alpha=1)
            x_ann = max(wl)
        plt.annotate(line.name, xy=(x_ann, Y_TOP-.2), rotation=-90, color=COLOR_TEXT)


    # # coverages
    #   =========
    HEIGHT = .7
    y = NUM_SLOTS-(1-HEIGHT)/2-1
    for coverage in cc:
        for interval in coverage.l0lf:
            x0, x1 = interval
            plt.plot([x0, x1, x1, x0, x0], [y, y, y-HEIGHT, y-HEIGHT, y], color=COLOR_COVERAGE*.5, alpha=.8)
            plt.fill_between(interval, [y, y], [y-HEIGHT, y-HEIGHT], color=COLOR_COVERAGE, alpha=.6, hatch='//')
            x_ann = interval[1]
        plt.annotate(coverage.name, xy=(x_ann+40, y-HEIGHT/2), color=COLOR_TEXT, verticalalignment="center")
        y -= 1

    plt.xlim([l0, lf])
    plt.ylim([0, NUM_SLOTS+.01])
    plt.xlabel("Wavelength ($\AA$)")
    plt.tight_layout()



class RedshiftWindow(QMainWindow):
    def __init__(self, sp_sky):
        QMainWindow.__init__(self)

        self._refs = []
        def keep_ref(obj):
            self._refs.append(obj)
            return obj

        self.sp_sky = sp_sky

        lw1 = keep_ref(QVBoxLayout())

        lwset = keep_ref(QHBoxLayout())
        lw1.addLayout(lwset)
        ###
        laa = keep_ref(QLabel("&Redshift"))
        lwset.addWidget(laa)
        ###
        sbx = self.spinBox_redshift = QDoubleSpinBox()
        laa.setBuddy(sbx)
        sbx.setSingleStep(.01)
        sbx.setDecimals(2)
        sbx.setMinimum(-.5)
        sbx.setMaximum(17)
        # sbx.valueChanged.connect(self.spinBoxValueChanged)
        lwset.addWidget(sbx)
        ###
        b = keep_ref(QPushButton("Redra&w"))
        lwset.addWidget(b)
        b.clicked.connect(self.redraw)
        ###
        lwset.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))


        ###
        wm = keep_ref(QWidget())
        # a99.set_margin(wm, 0)
        lw1.addWidget(wm)
        self.figure, self.canvas, self.lfig = a99.get_matplotlib_layout(wm)

        cw = self.centralWidget = QWidget()
        cw.setLayout(lw1)
        self.setCentralWidget(cw)
        self.setWindowTitle("MOSAIC Wavelength Chart")

        self.redraw()

    def get_redshift(self):
        return float(self.spinBox_redshift.value())


    # def spinBoxValueChanged(self, *args):
    #     self.label_redshiftValue.setText(str(self.get_redshift()))


    def redraw(self):
        try:
            fig = self.figure
            fig.clear()
            draw(fig, self.sp_sky, self.get_redshift())
            self.canvas.draw()
        except Exception as E:
            a99.get_python_logger().exception("Could not draw figure")



if __name__ == "__main__":

    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
    )
    parser.add_argument('--plot', action='store_const',
                        const=True, default=False,
                        help='Plot mode (default is GUI mode)')
    args = parser.parse_args()

    sp_sky = None
    try:
        sp_sky = ao.load_eso_sky()
    except:
        a99.get_python_logger().exception("Failed to load ESO sky model")

    if args.plot:
        fig = plt.figure()
        # draw(fig, telluric_spectra)
        draw(fig, sp_sky)
        plt.show()
    else:
        app = a99.get_QApplication([])
        form = RedshiftWindow(sp_sky)  # telluric_spectra)
        form.show()
        sys.exit(app.exec_())



_all_ = ["SpectrumBlock", "ToScalar", "SpectrumListBlock", "GroupBlock"]


import numpy as np
import copy
import f311.filetypes as ft


########################################################################################################################

class _BaseBlock(object):
    def __init__(self):
        self._input = None

    def use(self, inp):
        self._input = inp
        try:
            output = self._do_use(inp)
            return output
        finally:
            self._input = None

    def _do_use(self, inp):
        raise NotImplementedError()


class _SpectrumBlock(_BaseBlock):
    """Input is Spectrum (abbreviated "SB")"""

    def use(self, inp):
        assert isinstance(inp, ft.Spectrum)
        self._input = inp
        try:
            # If the output wavelength vector is the same, flag_copy_wavelength determines whether this vector
            # will be copied or just assigned to the output
            #
            # **Attention** blocks that handle the wavelength vectors must comply
            output = self._do_use(inp)
            assert output is not None  # it is common to forger to return in _do_use()
            if isinstance(output, ft.Spectrum):
                assert output._flag_created_by_block
            # Automatically assigns output wavelength vector if applicable
            if isinstance(output, ft.Spectrum) and output.wavelength is None and len(output.y) == len(inp.y):
                output.wavelength = np.copy(inp.wavelength)  # TODO this may slow down things... or not ... if self.flag_copy_wavelength else input.wavelength
            return output
        finally:
            self._input = None


class SpectrumBlock(_SpectrumBlock):
    """Spectrum-To-Spectrum"""

    def use(self, inp):
        assert isinstance(inp, ft.Spectrum)
        self._input = inp
        try:
            # If the output wavelength vector is the same, flag_copy_wavelength determines whether this vector
            # will be copied or just assigned to the output
            #
            # **Attention** blocks that handle the wavelength vectors must comply
            output = self._do_use(inp)
            assert output is not None  # it is common to forger to return in _do_use()
            assert isinstance(output, ft.Spectrum)
            assert output._flag_created_by_block
            # Automatically assigns output wavelength vector if applicable
            if isinstance(output, ft.Spectrum) and output.wavelength is None and len(output.y) == len(inp.y):
                output.wavelength = np.copy(inp.wavelength)  # TODO this may slow down things... or not ... if self.flag_copy_wavelength else input.wavelength
            return output
        finally:
            self._input = None

    def _new_output(self):
        """Call from _do_use() to create new spectrum based on input spectrum.

        Never create spectrum directly because we want to keep certain attributes, such as more_headers"""
        output = ft.Spectrum()
        output._flag_created_by_block = True  # assertion
        output.more_headers = copy.deepcopy(self._input.more_headers)
        return output

    def _copy_input(self, inp):
        """
        Returns copy of input + extra annotations

        Don't copy spectra explicitly; use this method instead
        """

        output = copy.deepcopy(inp)
        output._flag_created_by_block = True

        return output

    def _do_use(self, inp):
        """Default implementation is identity"""
        return self._copy_input(inp)


class ToScalar(_SpectrumBlock):
    """Ancestor for Spectrum Blocks whose output is a scalar (abbreviated "SpToSca")"""


class _SpectrumListBlock(_BaseBlock):
    """SpectrumBlock -- class with a "use" method accepting a SpectrumList as input"""

    def use(self, inp):
        assert isinstance(inp, ft.SpectrumList)
        self._input = inp
        try:
            output = self._do_use(inp)
            assert output is not None
            if isinstance(output, ft.SpectrumList):
                assert output._flag_created_by_block
            return output
        finally:
            self._input = None

    def _new_output(self):
        """Call from _do_use() to create new SpectrumList based on input"""
        output = ft.SpectrumList()
        output._flag_created_by_block = True  # assertion
        return output

    def _do_use(self, inp):
        """Default implementation is identity"""
        return inp


class SpectrumListBlock(_SpectrumListBlock):
    """Blocks whose output have the same number of spectra as the input (abbreviated "SpectrumListBlock")"""


class GroupBlock(_SpectrumListBlock):
    """
    Blocks whose output have only one spectrum (abbreviated "MDB")

    The "merge down" idea is that output from these blocks have only one spectrum
    (still stored as a spectrum list though)

    This class has no gear and exists for grouping purposes only
    """



__all__ = ["XPlotXY"]


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import numpy as np
import matplotlib.pyplot as plt
import a99
import f311.filetypes as ft


LINE_COLOR = 'k'
LINE_WIDTH = 2


class XPlotXY(a99.XLogMainWindow):
    """
    Plots two fields of a SpectrumCollection object in a simple x-y plot

    Args:
      collection: SpectrumCollection object
    """
    def __init__(self, collection, *args):
        a99.XLogMainWindow.__init__(self, *args)

        self._refs = []
        self._drawers = [self._draw_line, self._draw_error]
        self._buttons = []
        def keep_ref(obj):
            self._refs.append(obj)
            return obj

        assert isinstance(collection, ft.SpectrumCollection)
        self.collection = collection

        lw1 = keep_ref(QVBoxLayout())


        # First bar of plot options

        lwset0 = keep_ref(QHBoxLayout())
        lw1.addLayout(lwset0)
        ## Radio buttons for plot type
        # Radio button group probably needs its own layout, I think
        lwset0.addWidget(keep_ref(QLabel("Plot type")))
        lwset01 = QHBoxLayout()
        lwset0.addLayout(lwset01)
        r = self.rb_line = QRadioButton("Line")
        r.setChecked(True)
        self._buttons.append(r)
        lwset01.addWidget(r)
        r = self.rb_error = QRadioButton("Error bars")
        r.setToolTip("Groupx points by x-value and plots error bars "
                     "(center: mean; bar: +- standard deviation")
        self._buttons.append(r)
        lwset01.addWidget(r)
        lwset0.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))

        # Second bar of plot options

        lwset1 = keep_ref(QHBoxLayout())
        lw1.addLayout(lwset1)
        ###
        laa = keep_ref(QLabel("&X-axis"))
        lwset1.addWidget(laa)
        ###
        cbx = self.comboBoxX = QComboBox()
        cbx.addItems(collection.fieldnames)
        laa.setBuddy(cbx)
        lwset1.addWidget(cbx)
        ###
        laa = keep_ref(QLabel("&Y-axis"))
        lwset1.addWidget(laa)
        ###
        cby = self.comboBoxY = QComboBox()
        cby.addItems(collection.fieldnames)
        laa.setBuddy(cby)
        lwset1.addWidget(cby)
        ###
        b = keep_ref(QPushButton("Re&draw"))
        lwset1.addWidget(b)
        b.clicked.connect(self.redraw)
        ###
        lwset1.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))


        ###
        wm = keep_ref(QWidget())
        # a99.set_margin(wm, 0)
        lw1.addWidget(wm)
        self.figure, self.canvas, self.lfig = a99.get_matplotlib_layout(wm)

        cw = self.centralWidget = QWidget()
        cw.setLayout(lw1)
        self.setCentralWidget(cw)
        self.setWindowTitle("Plot Two Spectrum Collection Fields as a X-Y Chart")

        # self.redraw()

    # def get_redshift(self):
    #     return float(self.spinBox_redshift.value())


    # def spinBoxValueChanged(self, *args):
    #     self.label_redshiftValue.setText(str(self.get_redshift()))


    def redraw(self):
        try:
            idx = self._get_plot_type_index()
            if idx == -1:
                raise RuntimeError("Cannot draw, plot type not selected")

            fig = self.figure
            fig.clear()

            fieldname_x = str(self.comboBoxX.currentText())
            fieldname_y = str(self.comboBoxY.currentText())

            spectra = self.collection.spectra
            n = len(spectra)
            xx = np.zeros((n,))
            yy = np.zeros((n,))
            for i, sp in enumerate(spectra):
                s_x = sp.more_headers.get(fieldname_x)
                s_y = sp.more_headers.get(fieldname_y)

                # Will ignore error converting to float
                # Cells giving error will be zero
                try:
                    xx[i] = float(s_x)
                except TypeError:
                    pass
                try:
                    yy[i] = float(s_y)
                except TypeError:
                    pass

            sort_idxs = np.argsort(xx)

            xx = xx[sort_idxs]
            yy = yy[sort_idxs]

            self._drawers[idx](xx, yy)

            # Final adjustments
            k = 0.02 # percentual margin to the left and to the right
            xmin, xmax = np.min(xx), np.max(xx)
            xspan = xmax-xmin
            margin = .5 if xspan == 0 else xspan*k
            plt.xlim([xmin-margin, xmax+margin])
            plt.xlabel(fieldname_x)
            plt.ylabel(fieldname_y)
            plt.setp(list(plt.gca().spines.values()), linewidth=2)

            a99.format_BLB()
            self.canvas.draw()

        except Exception as e:
            a99.get_python_logger().exception("Error plotting")
            self.add_log_error("Could draw figure: "+a99.str_exc(e), True)


    def _get_plot_type_index(self):
        for i, b in enumerate(self._buttons):
            if b.isChecked():
                return i
        return -1


    def _draw_line(self, xx, yy):
        plt.plot(xx, yy, lw=2, color=LINE_COLOR)

    def _draw_error(self, xx, yy):
        """Draws error plot using x-values to group y-values"""

        xunique = np.unique(xx)
        n = len(xunique)
        y, errors = np.zeros(n), np.zeros(n)

        for i, xnow in enumerate(xunique):
            ysel = yy[xx == xnow]
            y[i] = np.mean(ysel)
            errors[i] = np.std(ysel)

        plt.plot(xunique, y, color=LINE_COLOR, lw=LINE_WIDTH, linestyle="--")
        plt.errorbar(xunique, y, yerr=errors, fmt="o", markeredgewidth=LINE_WIDTH, markersize=8, markerfacecolor=(1., 1., 1.), color=LINE_COLOR, lw=LINE_WIDTH)



from .spcollection import *
from .fullcube import *
from .sparsecube import *
from .splist import *
from .filefullcube import *
from .filepar import *
from .filesparsecube import *
from .filesplist import *


"""
Basic routines, they don't use anything from parent module
"""


import numpy as np
import f311.physics as ph

__all__ = ["sparse_cube_to_colors", "spectrum_to_rgb"]

def sparse_cube_to_colors(scube, visible_range=None, flag_scale=False, method=0):
    """Returns a [nY, nX, 3] red-green-blue (0.-1.) matrix

    Args:
      visible_range=None: if passed, the true human visible range will be
                            affine-transformed to visible_range in order
                            to use the red-to-blue scale to paint the pixels
      flag_scale: whether to scale the luminosities proportionally
                    the weight for each spectra will be the area under the flux
      method: see f311.physics.spectrum_to_rgb()
    """
    im = np.zeros((scube.height, scube.width, 3))
    weights = np.zeros((scube.height, scube.width, 3))
    max_area = 0.
    for i in range(scube.width):
        for j in range(scube.height):
            sp = scube.get_pixel(i, j, False)
            if len(sp) > 0:
                im[j, i, :] = ph.spectrum_to_rgb(sp, visible_range, method)
                sp_area = np.sum(sp.y)
                max_area = max(max_area, sp_area)
                if flag_scale:
                    weights[j, i, :] = sp_area
    if flag_scale:
        weights *= 1. / max_area
        im *= weights
    return im


def spectrum_to_rgb(sp, visible_range=None, method=0):
    """Takes weighted average of rainbow colors RGB's

    Args:
        sp: f311.filetypes.Spectrum instance
        visible_range=None: if passed, affine-transforms the rainbow colors
        method --
          0: rainbow colors
          1: RGB
    """

    if len(visible_range) < 2:
        raise RuntimeError("Invalid visible range: {0!s}".format(visible_range))
    if visible_range[1] <= visible_range[0]:
        raise RuntimeError(
            "Second element of visible range ({0!s}) must be greater than first element".format(
                visible_range))

    if method == 1:
        # new color system splitting visible range in three
        dl = float(visible_range[1] - visible_range[0]) / 3
        ranges = np.array(
            [[visible_range[0] + i * dl, visible_range[0] + (i + 1) * dl] for i in range(3)])
        colors = np.array([(0., 0., 1.), (0., 1., 0.), (1., 0., 0.)])  # blue, green, red

        tot_area, tot_sum = 0., np.zeros(3)
        for color, range_ in zip(colors, ranges):
            b = np.logical_and(sp.x >= range_[0], sp.x <= range_[1])
            area = np.sum(sp.y[b])
            tot_area += area
            tot_sum += color * area
        if tot_area == 0.:
            tot_area = 1.
        ret = tot_sum / tot_area
        return ret

    elif method == 0:
        tot_area, tot_sum = 0., np.zeros(3)

        def ftrans(x):
            return x

        if visible_range:
            ll0 = ph.rainbow_colors[0].l0
            llf = ph.rainbow_colors[-1].lf
            ftrans = lambda lold: visible_range[0] + (visible_range[1] - visible_range[0]) / (
            llf - ll0) * (lold - ll0)

        for color in ph.rainbow_colors:
            b = np.logical_and(sp.x >= ftrans(color.l0), sp.x <= ftrans(color.lf))
            area = np.sum(sp.y[b])
            tot_area += area
            tot_sum += color.rgb * area
        if tot_area == 0.:
            tot_area = 1.
        ret = tot_sum / tot_area
        return ret
    else:
        raise RuntimeError("Unknown method: {0!s}".format(method))



from .. import FilePy
import a99
from ..molconsts import MolConsts

__all__ = ["FileMolConsts"]

@a99.froze_it
class FileMolConsts(FilePy):
    """Python source containing 'fobj = MolConsts(...)"""

    description = "molecular constants"
    default_filename = "configmolconsts.py"
    attrs = ["molconsts"]
    editors = ["mced.py"]

    def __init__(self):
        FilePy.__init__(self)

        self.molconsts = MolConsts()

    def _do_load(self, filename):
        module = a99.import_module(filename)
        self._copy_attr(module, "molconsts", MolConsts)

    def _do_save_as(self, filename):
        with open(filename, "w") as h:
            h.write("{}\n"
                    "from f311.filetypes import MolConsts\n"
                    "\n"
                    "molconsts = {}\n".format(self._get_header(), self.molconsts))



#!/usr/bin/env python

"""
Create several .splist (spectrum list) files from WebSim-COMPASS output files; groups spectra that share same wavelength vector

All spectra in each .splist file will have the same wavelength vector
"""


import argparse
import logging
from f311 import aosss
import a99


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
    )

    parser.add_argument('--stage', type=str, nargs='?', default="spintg",
     help="Websim-Compass pipeline stage (will collect files named, e.g., C000793_<stage>.fits)")

    args = parser.parse_args()

    aosss.create_spectrum_lists(".", args.stage)






"""
Spectrograph modes specifications
"""

from a99 import AttrsPart


class SpectrographMode(AttrsPart):
    """
    Spectrograph mode specification

    Args:
        name: name of spectrograph mode
        abbreviation: abbreviation of spectrograph mode
        aperture: aperture in (arcsec)
        multiplex: maximum number of simultaneous objects being observed (unitless)
        wavelength_range: [lambda_min, lambda_max] (angstrom)
        resolving_power: lambda/delta_lambda (unitless)
        fov: field-of-view (arcmin**2)
        flag_glao: whether or not Ground-Layer Adaptive Optics is available for this mode
        flag_moao: whether of not Multi-Object Adaptive Optics is available for this mode
        pixel_size: spatial pixel size
    """

    attrs = ["name", "abbreviation", "aperture", "multiplex", "wavelength_range", "resolving_power",
             "fov", "flag_glao", "flag_moao", "pixel_size"]

    def __init__(self, name="", abbreviation="", aperture=0., multiplex=0.,
                 wavelength_range=(0., 0.), resolving_power=0., fov=0., flag_glao=False,
                 flag_moao=False, pixel_size=0.):
        AttrsPart.__init__(self)

        self.name = name
        self.abbreviation = abbreviation
        self.aperture = aperture
        self.multiplex = multiplex
        self.wavelength_range = wavelength_range
        self.resolving_power = resolving_power
        self.fov = fov
        self.flag_glao = flag_glao
        self.flag_moao = flag_moao
        self.pixel_size = pixel_size



modes = [
    SpectrographMode(name="High Multiplex Mode - Visible",
                     abbreviation="HMM-Vis",
                     aperture=.9,
                     multiplex=200,
                     wavelength_range=(4000., 7000.),
                     resolving_power=15000.,
                     fov=40.,
                     flag_glao=True,
                     flag_moao=False
                    ),
    SpectrographMode(name="High Multiplex Mode - NIR",
                     abbreviation="HMM-NIR",
                     aperture=.6,
                     multiplex=200,
                     wavelength_range=(7000., 18000.),
                     resolving_power=5000.,
                     fov=40.,
                     flag_glao=True,
                     flag_moao=False
                     ),
    SpectrographMode(name="High Definition Mode (essential)",
                     abbreviation="HDM (essential)",
                     aperture=2.,
                     multiplex=10,
                     wavelength_range=(10000., 18000.),
                     resolving_power=5000.,
                     fov=40.,
                     flag_glao=True,
                     flag_moao=True,
                     pixel_size=80.,
                     ),
    SpectrographMode(name="High Definition Mode (desirable)",
                     abbreviation="HDM (desirable)",
                     aperture=2.,
                     multiplex=10,
                     wavelength_range=(8000., 25000.),
                     resolving_power=15000.,
                     fov=40.,
                     flag_glao=True,
                     flag_moao=True,
                     pixel_size=40.,
                     ),
    SpectrographMode(name="Inter-Galactic Mode (essential)",
                     abbreviation="IGM (essential)",
                     aperture=3.,
                     multiplex=30,
                     wavelength_range=(4000., 8000.),
                     resolving_power=3000.,
                     fov=40.,
                     flag_glao=True,
                     flag_moao=False,  # TODO not sure if MOAO is really not available in IGM
                     pixel_size=250.,
                     ),
    SpectrographMode(name="Inter-Galactic Mode (desirable)",
                     abbreviation="IGM (desirable)",
                     aperture=3.,
                     multiplex=30,
                     wavelength_range=(3700., 10000.),
                     resolving_power=5000.,
                     fov=40.,
                     flag_glao=True,
                     flag_moao=False,  # TODO not sure if MOAO is really not available in IGM
                     pixel_size=250.,
                     ),
    ]




__all__ = ["FileSparseCube"]


import a99
from .. import DataFile
from . import SparseCube
from astropy.io import fits


@a99.froze_it
class FileSparseCube(DataFile):
    """FITS Sparse Data Cube (storage to take less disk space)"""
    attrs = ['sparsecube']
    description = "Data Cube (FITS file)"
    default_filename = "default.sparsecube"
    flag_txt = False

    def __init__(self):
        DataFile.__init__(self)
        self.sparsecube = SparseCube()

    def _do_load(self, filename):
        fits_obj = fits.open(filename)
        self.sparsecube = SparseCube()
        self.sparsecube.from_hdulist(fits_obj)
        self.filename = filename

    def _do_save_as(self, filename):
        hdul = self.sparsecube.to_hdulist()
        a99.overwrite_fits(hdul, filename)

    def init_default(self):
        # Already created OK
        pass



#!/usr/bin/env python
"""
Tunes the "zinf" parameter for each atomic line in atomic lines file

The "zinf" parameter is a distance in angstrom from the centre of an atomic
line. It specifies the calculation range for the line:
[centre-zinf, centre+zinf].

This script runs pfant for each atomic line to determine the width of each
atomic line and thus zinf.

Note: pfant is run using most of its default settings and will require the
following files to exist in the current directory:
  - main.dat
  - dissoc.dat
  - abonds.dat
  - modeles.mod
  - partit.dat
  - absoru2.dat

Note: the precision in the zinf found depends on the calculation step ("pas")
specified in main.dat. A higher "pas" means lower precision and a tendency to
get higher zinf's. This is really not critical. pas=0.02 or pas=0.04 should do.

"""

import argparse
import logging
import copy
import numpy as np
import os.path
import a99
import f311.filetypes as ft
import f311.explorer as ex
import f311.filetypes as ft
import f311.explorer as ex


a99.logging_level = logging.INFO
a99.flag_log_file = True
a99.flag_log_console = True

# Fraction of peak to be considered its "settlement"
EPSILON = 1e-4

DEFOUT = "<made-up filename>"


def _get_zinf(lambda_centre, norm):
    """Returns zinf given *normalized* spectrum, or 0 if line if flat.

    Args:
      lambda_centre: centre of line
      norm: Spectrum instance

    Spectrum is expected to contain one line only.
    """

    y = norm.y  # normalized spectrum: values between 0 and 1
    if np.all(np.diff(y) == 0):
        # flat line
        return 0
    ymax = 1
    ymin = min(y)
    i_zinf = np.argmax(norm.y < ymax-(ymax-ymin)*EPSILON)
    zinf = lambda_centre-norm.x[i_zinf]
    return zinf


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
     )
    parser.add_argument('--min', type=float, nargs='?', default=.1,
     help='minimum zinf. If zinf found for a particular line is smaller than '
          'this value, this value will be used instead')
    parser.add_argument('--max', type=float, nargs='?', default=50.,
     help='maximum zinf. If zinf found for a particular line is greater than '
          'this value, this value will be used instead')
    parser.add_argument('--inflate', type=float, nargs='?', default=1.1,
     help='Multiplicative constant to apply a "safety margin". '
          'Each zinf found will be multiplied by this value. For example '
          'a value of INFLATE=1.1 means that all the zinf\'s saved will be 10 '
          'percent larger than those calculated')
    parser.add_argument('--ge_current', action="store_true",
     help='"Greater or Equal to current": If this option is set, the current '
          'zinf in the atomic lines file is used as a lower boundary.')
    parser.add_argument('fn_input', type=str, help='input file name', nargs=1)
    parser.add_argument('fn_output', type=str, help='output file name',
     nargs="?", default=DEFOUT)
    parser.add_argument('--no_clean', action="store_true",
    help='If set, will not remove the session directories.')

    args = parser.parse_args()
    logger = a99.get_python_logger()

    fn_out = args.fn_output
    if fn_out == DEFOUT:
        temp = args.fn_input[0].replace("atoms-untuned-", "")
        fn_out = "atoms-"+temp
        logger.info("Made up output filename: '%s'" % fn_out)

    logger.info("Using minimum zinf = %g" % args.min)
    logger.info("Using MAXIMUM zinf = %g" % args.max)
    logger.info("Using inflate = %g" % args.inflate)
    logger.info("Using ge_current = %s" % args.ge_current)

    file_atoms = ft.FileAtoms()
    file_atoms.load(args.fn_input[0])

    logger.info("Number of lines in file '%s': %d" % \
     (args.fn_input[0], file_atoms.num_lines))

    # # Tries to get input files together
    pf.setup_inputs(h=False, atoms=False, molecules=False, opa=False)


    # ## Runs innewmarcs if modeles.mod not present
    
    if not os.path.isfile('modeles.mod'):
        logger.info("'modeles.mod' does not exist, running innewmarcs...")
        inn = pf.Innewmarcs()
        inn.conf.opt.opa = False
        inn.run()



    # ## As suggested by M. Trevisan, adds 0.5 to all abundances, except ...
    K_ADD = 0.5
    EXCEPTIONS = ["FE", "HE"]
    logger.info("Preparing abundances (Adding %g to all abundances, " % K_ADD)
    logger.info("except those of %s)..." % str(EXCEPTIONS))
    fa = ft.FileAbonds()
    fa.load("abonds.dat")
    cnt = 0
    for i in range(len(fa)):
       if fa.ele[i] not in EXCEPTIONS:
           fa.abol[i] += K_ADD
           cnt += 1
    logger.info("%d abundance(s) changed" % cnt)


    # This is old code, with a minimum value for the abundances but I don't think it is very reallistic
    # MIN_ABOND = 3
    # logger.info("Preparing abundances file with all abundances >= %g ..." % MIN_ABOND)
    # fa = FileAbonds()
    # fa.load("abonds.dat")
    # cnt = 0
    # for i in range(len(fa)):
    #     if fa.abol[i] < MIN_ABOND:
    #         fa.abol[i] = MIN_ABOND
    #         cnt += 1
    # logger.info("%d abundance(s) changed" % cnt)


    # ## Creates the Runnable instances
    PAS = 0.04
    logger.info("Calculation step will be set to %g" % PAS)
    logger.info("Preparing pfant's...")
    combos, ll, i = [], [], 0
    ii = 0
    n = file_atoms.num_lines
    for atom in file_atoms.atoms:
        for line in atom.lines:
            a = copy.copy(atom)  # shallow copy
            a.lines = [line]
            f = ft.FileAtoms()
            f.atoms = [a]
            combo = pf.Combo([pf.FOR_PFANT])
            # Fortran messages will not be displayed in terminal
            combo.conf.flag_log_console = False
            combo.conf.sid.flag_split_dirs = True
            # Forces pfant to run one single iteration (ikeytot will be =1)
            combo.conf.opt.aint = args.max+10
            combo.conf.file_atoms = f
            combo.conf.file_abonds = fa
            combo.conf.opt.logging_level = "warning"
            combo.conf.opt.pas = PAS
            combo.conf.flag_output_to_dir = True
            combo.conf.opt.zinf = args.max
            combo.conf.opt.no_molecules = True
            combo.conf.opt.opa = False
            combo.conf.opt.no_h = True
            # Note that half of the line (needs to)/(will be) calculated
            combo.conf.opt.llzero = line.lambda_-args.max
            combo.conf.opt.llfin = line.lambda_
            combos.append(combo)
            ll.append(a)

            i += 1
            ii += 1
            if ii == 1000:
                print(a99.format_progress(i, n))
                ii = 0

    # # Runs pfant
    logger.info("Running pfant's...")






    # FLAG_CLEAN = True
    # rm = RunnableManager(flag_auto_clean=FLAG_CLEAN)






    rm = pf.RunnableManager()
    rm.add_runnables(combos)
    app = a99.get_QApplication([])
    form = pf.XRunnableManager(None, rm)
    form.show()
    # it is good to start the manager as late as possible, otherwise
    # the program will hang if, for example, the form fails to be created.
    rm.start()
    app.exec_()
    rm.exit()

    # # Saves log
    LOG_FILENAME = "tune-zinf-status.log"
    with open(LOG_FILENAME, "w") as h:
      h.write(str(rm))
    logger.info("Final status saved to file '%s'" % LOG_FILENAME)

    # # Finds zinf's and creates output file
    flag_success, flag_clean = False, False
    if not rm.flag_finished:
        logger.info("Not finished")
    elif rm.num_failed > 0:
        logger.info("Number of failed tasks: %d" % rm.num_failed)
        logger.info("Please check log files inside directories of sessions that failed")
        logger.info("(session directories will not be removed).")
    else:
        logger.info("Finding zinf's, please wait...")
        # # Calculates zinf and save new atomic lines file
        n = len(combos)
        X = np.zeros((n, 3))  # [algf, kiex, zinf], ...]
        ii = 0
        print(a99.format_progress(0, n))
        cnt_min = 0
        cnt_max = 0
        for i, (a, combo) in enumerate(zip(ll, combos)):
            combo.pfant.load_result()


            # if not FLAG_CLEAN:
            #     combo.pfant.load_result()
            # else:
            #     pass  # result will be already loaded



            norm = combo.pfant.norm  # normalized spectrum
            line = a.lines[0]  # a is single-line FileAtoms object
            zinf = _get_zinf(line.lambda_, norm)
            zinf *= args.inflate
            if zinf == 0:
                # Note that some lines don't appear at all, so there is no way
                # to determine zinf
                logger.warning("zinf is ZERO: (%s) (%s) (%s)" %
                 (combo.conf.sid.dir, a.one_liner_str(), line.one_liner_str()))
            if zinf < args.min:
                zinf = args.min
                cnt_min += 1
            if zinf > args.max:
                # print " %gOLHOLHO (%s) (%s) (%s)" % \
                #  (zinf, combo.conf.session_dir, a.one_liner_str(), line.one_liner_str())
                zinf = args.max
                cnt_max += 1
            if args.ge_current and line.zinf > zinf:
                zinf = line.zinf
            line.zinf = zinf

            i += 1
            ii += 1
            if ii == 100:
                print(a99.format_progress(i, n))
                ii = 0
        print(a99.format_progress(n, n))
        logger.info("zinf's clipped to minimum: %d" % cnt_min)
        logger.info("zinf's clipped to maximum: %d" % cnt_max)

        logger.info("Creating output file...")
        file_atoms.save_as(fn_out)
        flag_clean = not args.no_clean
        flag_success = True


    # # Removes session-* directories
    if flag_clean:
        # this part removes sessionxxxx/sessionyyyy
        logger.info("Cleaning...")

        for i, combo in enumerate(combos):
            try:
                combo.sid.clean()
            except Exception as E:
                logger.info("Error cleaning session for %s: %s" % (combo.name, str(E)))
        # this part removes sessionxxxxx
        # but also removes previous sessions
        # not gonna do it dd = glob.glob("session*")

    if flag_success:
        logger.info("Successfully created file '%s' !!" % fn_out)



"""Source files for the Graphical User Interface applications in this package"""

from .a_XText import *
from .a_XExplorer import *
from .gui_pyfant import *
from .gui_aosss import *
from .a_XFileMainWindowBase import *
from .a_XFileMainWindow import *
from .a_WChooseSpectrum import *
from .gui_convmol import *


__all__ = ["WSpectrumCollectionBase"]

import copy
import os
import os.path
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from .a_XScaleSpectrum import *
from astropy import units as u
import a99
# from .... import explorer as ex
import f311.filetypes as ft


# TODO options window to set this up
_CONFIG_LEGEND = "/gui/WSpectrumCollectionBase/plot_overlapped/flag_legend"


class WSpectrumCollectionBase(a99.WEditor):
    """Base class for WSpectrumList and WSparseCube

    This class implements some operations that are the same for either WSpectrumList and WSparseCube

    This class does not implement the visual stuff.
    """


    def __init__(self, parent, flag_splist=True):
        a99.WEditor.__init__(self, parent)

        def keep_ref(obj):
            self._refs.append(obj)
            return obj

        self.collection = None # SpectrumCollection
        self.flag_splist = flag_splist


        # # Creates actions

        action = self.action_add_spectra = QAction(a99.get_icon("list-add"), "&Add spectra...", self)
        action.setToolTip("Opens a 'Open File' window where multiple files can be selected.\n"
                          "Accepts any supported spectral type, and also other Spectrum Collection files.\n"
                          "Files are added in alphabetical order.")
        action.triggered.connect(self.on_add_spectra)

        action = self.action_all_to_scalar = QAction(a99.get_icon("go-next"), "To &Scalar...", self)
        action.triggered.connect(self.on_all_to_scalar)

        action = self.action_all_plot_xy = QAction(a99.get_icon("visualization"), "X-&Y Plot", self)
        action.triggered.connect(self.on_plot_xy)

        action = self.action_all_plot_xyz = QAction(a99.get_icon("visualization"), "X-Y-&Z Plot", self)
        action.triggered.connect(self.on_plot_xyz)

        action = self.action_all_export_csv = QAction(a99.get_icon("document-export"), "&Export CSV...", self)
        action.triggered.connect(self.on_all_export_csv)

        action = self.action_sel_use_spectrum_block = QAction(a99.get_icon("go-next"), "&Transform...", self)
        action.setToolTip("Selected spectra will be deleted and transformed spectra will be added at the end")
        action.triggered.connect(self.on_sel_use_spectrum_block)

        action = self.action_sel_plot_stacked = QAction(a99.get_icon("chart2"), "Plot &Stacked", self)
        action.triggered.connect(self.on_sel_plot_stacked)

        action = self.action_sel_plot_overlapped = QAction(a99.get_icon("chart1"), "Plot &Overlapped", self)
        action.triggered.connect(self.on_sel_plot_overlapped)

        action = self.action_sel_open_in_new = QAction(a99.get_icon("window-new"), "Open in new window", self)
        action.triggered.connect(self.on_sel_open_in_new)

        action = self.action_sel_delete = QAction(a99.get_icon("list-remove"), "Delete", self)
        action.triggered.connect(self.on_sel_delete)

        action = self.action_curr_scale = QAction(a99.get_icon("zoom-fit"), "Scale to Magnitude...", self)
        action.triggered.connect(self.on_curr_scale)



    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Interface

    def set_collection(self, x):
        assert isinstance(x, ft.SpectrumCollection)
        self.collection = x
        self._update_gui()
        self.setEnabled(True)


    def get_selected_spectra(self):
        return [self.collection.spectra[i] for i in self.get_selected_spectrum_indexes()]


    def get_selected_row_indexes(self):
        ii = list(set([index.row() for index in self.twSpectra.selectedIndexes()]))
        return ii


    def get_selected_spectrum_indexes(self):
        items = self.twSpectra.selectedItems()
        ii = []
        for item in items:
            obj = item.data(1)
            if isinstance(obj, int):
                ii.append(obj)
        ii.sort()
        return ii


    def row_index_to_spectrum_index(self, row_index):
        """
        Converts table row index within self.collection.spectra

        This is necessary because the table may be sorted
        """

        item = self.twSpectra.item(row_index, self.twSpectra.columnCount()-1)
        ret = item.data(1)
        return ret


    def get_current_spectrum_index(self):
        row_index = self.twSpectra.currentRow()
        if row_index == -1:
            return -1
        return self.row_index_to_spectrum_index(row_index)


    def get_current_spectrum(self):
        """Returns spectrum on which the table cursor (movable with the keyboar arrows) is currently"""
        spectrum_index = self.get_current_spectrum_index()
        if spectrum_index == -1:
            return None
        return self.collection.spectra[spectrum_index]

    def update(self):
        """Refreshes the GUI to reflect what is in self.collection"""
        self._update_gui()


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Qt override

    def setFocus(self, reason=None):
        """Sets focus to first field. Note: reason is ignored."""
        self.twSpectra.setFocus()


    def eventFilter(self, source, event):
        if event.type() == QEvent.FocusIn:
            # text = random_name()
            # self.__add_log(text)
            pass

        if event.type() == QEvent.KeyPress:
            if event.key() == Qt.Key_Delete:
                if source == self.twSpectra:
                    n_deleted = self._delete_spectra()
                    if n_deleted > 0:
                        self.changed.emit(False)
        return False

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Slots

    def section_moved(self, idx_logical, idx_vis_old, idx_vis_new):
        obj = self.collection
        l = self.collection.fieldnames_visible
        fieldname_current = l[idx_vis_old]
        del l[idx_vis_old]
        l.insert(idx_vis_new, fieldname_current)


    def show_header_context_menu(self, position):
        obj = self.collection

        ah = self.twSpectra.horizontalHeader()
        # col_idx = ah.logicalIndexAt(position)
        col_idx = ah.visualIndex(ah.logicalIndexAt(position))

        menu = QMenu()
        act_hide_current = None
        fieldname_current = None

        if col_idx < len(obj.fieldnames_visible):
            fieldname_current = obj.fieldnames_visible[col_idx]
            act_hide_current = menu.addAction("&Hide field '%s'" % fieldname_current)
            menu.addSeparator()

        act_show_all = menu.addAction("&Show all fields")
        act_hide_all = menu.addAction("&Hide all fields")
        act_restore_order = menu.addAction("&Restore order")
        menu.addSeparator()

        aa_visible = []
        for fieldname in obj.fieldnames:
            act = menu.addAction(fieldname)
            act.setCheckable(True)
            act.setChecked(fieldname in obj.fieldnames_visible)
            aa_visible.append(act)

        action = menu.exec_(self.twSpectra.mapToGlobal(position))
        flag_update = False
        if action == act_hide_current and fieldname_current is not None:
            obj.fieldnames_visible.remove(fieldname_current)
            flag_update = True

        elif action == act_show_all:
            for fieldname in reversed(obj.fieldnames):
                if not fieldname in obj.fieldnames_visible:
                    obj.fieldnames_visible.insert(0, fieldname)
                    flag_update = True

        elif action == act_hide_all:
            obj.fieldnames_visible = []
            flag_update = True

        elif action == act_restore_order:
            curr_visible = copy.copy(obj.fieldnames_visible)
            obj.fieldnames_visible = []
            for fieldname in obj.fieldnames:
                if fieldname in curr_visible:
                    obj.fieldnames_visible.append(fieldname)
            flag_update = not (curr_visible == obj.fieldnames_visible)

        elif action in aa_visible:
            idx = aa_visible.index(action)
            if not aa_visible[idx].isChecked():
                obj.fieldnames_visible.remove(obj.fieldnames[idx])
            else:
                obj.fieldnames_visible.insert(0, obj.fieldnames[idx])
            flag_update = True

        if flag_update:
            self._update_gui()
            self.changed.emit(False)

    def on_twSpectra_customContextMenuRequested(self, position):
        """Mounts, shows popupmenu for the tableWidget control, and takes action."""
        obj = self.collection
        menu = QMenu()
        act_del = menu.addAction("&Delete selected (Del)")

        action = menu.exec_(self.twSpectra.mapToGlobal(position))
        flag_update = False
        if action == act_del:
            n_deleted = self._delete_spectra()
            if n_deleted > 0:
                self.changed.emit(False)

        if flag_update:
            self.changed.emit(False)
            self._update_gui()

    def on_all_to_scalar(self):
        flag_emit, flag_changed_header = False, False

        from .a_XToScalar import XToScalar
        form = self.keep_ref(XToScalar())
        if not form.exec_():
            return

        # It sth fails, will restore original
        save = copy.deepcopy(self.collection)
        try:
            for sp in self.collection.spectra:
                sp.more_headers[form.fieldname] = form.block.use(sp)

            if not form.fieldname in self.collection.fieldnames:
                self.collection.fieldnames.insert(0, form.fieldname)

            if not form.fieldname in self.collection.fieldnames_visible:
                self.collection.fieldnames_visible.insert(0, form.fieldname)

            self._update_gui()
            flag_emit = True

        except Exception as E:
            # Restores and logs error
            self.add_log_error("Failed to extract scalar: %s" % str(E), True)
            self.collection = save
            raise

        if flag_emit:
            self.changed.emit(True)


    def on_plot_xy(self):
        # Import here to circumvent cyclic dependency
        from .a_XPlotXY import XPlotXY
        form = self.keep_ref(XPlotXY(self.collection))
        form.show()


    def on_plot_xyz(self):
        # Import here to circumvent cyclic dependency
        from .a_XPlotXYZ import XPlotXYZ
        form = self.keep_ref(XPlotXYZ(self.collection))
        form.show()


    def on_all_export_csv(self):
        new_filename = QFileDialog.getSaveFileName(self, "Export text file (CSV format)", "export.csv", "*.csv")[0]
        if new_filename:
            # self.save_dir, _ = os.path.split(str(new_filename))
            try:
                lines = self.collection.to_csv()
                with open(str(new_filename), "w") as file:
                    file.writelines(lines)
            except Exception as E:
                msg = str("Error exporting text file: %s" % a99.str_exc(E))
                self.add_log_error(msg, True)
                raise


    def on_sel_use_spectrum_block(self):
        flag_emit, flag_changed_header = False, False

        from .a_XUseSpectrumBlock import XUseSpectrumBlock
        form = self.keep_ref(XUseSpectrumBlock())
        if not form.exec_():
            return

        # It sth fails, will restore original
        save = copy.deepcopy(self.collection)
        try:
            sspp = self.get_selected_spectra()
            self.collection.delete_spectra(self.get_selected_spectrum_indexes())
            for sp in sspp:
                block = copy.deepcopy(form.block)
                self.collection.add_spectrum(block.use(sp))

            self._update_gui()
            flag_emit = True

        except Exception as E:
            # Restores and logs error
            self.collection = save
            self.add_log_error("Failed to transform spectra: %s" % str(E), True)
            raise

        if flag_emit:
            self.changed.emit(True)


    def on_sel_plot_stacked(self):
        from f311 import explorer as ex

        sspp = self.get_selected_spectra()
        if len(sspp) > 0:
            ex.plot_spectra(sspp)


    def on_sel_plot_overlapped(self):
        from f311 import explorer as ex

        sspp = self.get_selected_spectra()
        flag_legend = ex.get_config().get_item(_CONFIG_LEGEND, True)
        oo = ex.PlotSpectrumSetup(flag_legend=flag_legend)
        if len(sspp) > 0:
            ex.plot_spectra_overlapped(sspp, setup=oo)


    def on_sel_open_in_new(self):
        ii = self.get_selected_spectrum_indexes()
        if len(ii) > 0:
            other = copy.deepcopy(self.collection)
            other.spectra = [copy.deepcopy(other.spectra[i]) for i in ii]
            # TODO who said it is a spectrum list? Could well be a FileSparseCube!!!! How to sort this?????????????????????????????????????????????????????????????????????

            # "Gambiarra"
            if self.flag_splist:
                from .a_XFileSpectrumList import XFileSpectrumList
                f = ft.FileSpectrumList()
                form = self.keep_ref(XFileSpectrumList())
            else:
                from .a_XFileSparseCube import XFileSparseCube
                f = ft.FileSparseCube()
                f.sparsecube = other
                form = self.keep_ref(XFileSparseCube())

            # f.filename = "noname"

            form.load(f)
            form.show()


    def on_sel_delete(self):
        n = self._delete_spectra()
        if n > 0:
            self.changed.emit(False)


    def on_curr_scale(self):
        """Performs a scaling operation on the current spectrum"""
        sp = self.get_current_spectrum()

        if sp is None:
            return

        form = XScaleSpectrum()
        form.set_spectrum(sp)
        if form.exec_():
            k = form.factor()
            if k != 1:
                sp.y *= k
                self._update_gui()
                self.changed.emit(False)


    def on_twSpectra_cellChanged(self, row, column):
        """Cell has been changed manually: commit to self.collection"""
        if self.flag_process_changes:
            flag_emit = False
            text = None
            item = self.twSpectra.item(row, column)
            name = self._get_tw_header(column)
            self.flag_process_changes = False
            try:
                value = str(item.text())
                # Tries to convert to float, otherwise stores as string
                try:
                    value = float(value)
                except:
                    pass

                # Certain fields must evaluate to integer because they are pixel positions
                if name in ("PIXEL-X", "PIXEL-Y", "Z-START"):
                    value = int(value)

                # Units must be a valid astropy.Unit() argument
                if name in ("X-UNIT", "Y-UNIT"):
                    value = u.Unit(value)

                spectrum_index = self.row_index_to_spectrum_index(row)

                self.collection.spectra[spectrum_index].more_headers[name] = value

                flag_emit = True
                # replaces edited text with eventually cleaner version, e.g. decimals from integers are discarded
                item.setText(str(value))

            except Exception as E:
                # restores original value
                item.setText(str(self.collection.spectra[row].more_headers.get(name)))
                self.add_log_error(a99.str_exc(E), True)
                raise

            finally:
                self.flag_process_changes = True

            if flag_emit:
                self.changed.emit(False)


    def on_twSpectra_itemSelectionChanged(self):
        """Updates actions enabled state"""
        self._update_enabled_actions()


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Internal gear


    def _get_tw_header(self, column):
        return str(self.twSpectra.horizontalHeaderItem(column).text())


    def _update_gui(self):
        """Updates GUI to match self.collection, i.e., mainly the table widget"""

        self.flag_process_changes = False
        try:
            # Builds table widget contents
            spectra = self.collection.spectra
            t = self.twSpectra
            n = len(spectra)
            FIXED = ["Spectrum summary report"]
            fieldnames_visible = self.collection.fieldnames_visible
            all_headers = fieldnames_visible+FIXED
            nc = len(all_headers)
            a99.reset_table_widget(t, n, nc)
            t.setHorizontalHeaderLabels(all_headers)
            i = 0
            for sp in spectra:
                j = 0

                # Spectrum.more_headers columns
                for h in fieldnames_visible:
                    twi = QTableWidgetItem(str(sp.more_headers.get(h)))
                    if h in "Z-START":  # fields that should be made read-only
                        twi.setFlags(twi.flags() & ~Qt.ItemIsEditable)
                    t.setItem(i, j, twi)
                    j += 1

                # Spectrum self-generated report
                twi = QTableWidgetItem(sp.one_liner_str())
                twi.setFlags(twi.flags() & ~Qt.ItemIsEditable)
                # stores spectrum index not to lose track in case the table is sorted by column
                twi.setData(1, i)
                t.setItem(i, j, twi)
                j += 1

                i += 1

            t.resizeColumnsToContents()

        finally:
            self._update_enabled_actions()

            self.flag_process_changes = True

    def _delete_spectra(self):
        ii = self.get_selected_spectrum_indexes()
        if len(ii) > 0:
            self.collection.delete_spectra(ii)
            self._update_gui()

        return len(ii)



"""
"Collaboration" with other packages
"""


from collections import OrderedDict
import importlib
import a99

__all__ = ["classes_vis", "get_suitable_vis_classes"]


__COLLABORATORS = ["f311.explorer"]


def get_suitable_vis_classes(obj):
    """Retuns a list of Vis classes that can handle obj."""

    ret = []
    for class_ in classes_vis():
        if isinstance(obj, class_.input_classes):
            ret.append(class_)
    return ret


def get_suitable_vis_list_classes(objs):
    """Retuns a list of VisList classes that can handle a list of objects."""

    from f311 import explorer as ex

    ret = []
    for class_ in classes_vis():
        if isinstance(class_, ex.VisList):
            flag_can = True
            for obj in objs:
                if not isinstance(obj, class_.item_input_classes):
                    flag_can = False
                    break
            if flag_can:
                ret.append(class_)
    return ret


# def collaborators():
#     """Returns a dictionary of packages scanned to populate _classes_*
#
#     Example: {"f311.explorer": f311.explorer, }
#     """
#     if __flag_first:
#         __setup()
#     return __collaborators

def _collect_classes(m):
    """
    Adds entries to _classes_*

    Args:
        m: module object that must contain the following sub-modules: datatypes, vis
    """

    from f311 import explorer as ex

    def _extend(classes, newclasses):
        """Filters out classes already present in list.

        This shouldn't be necessary, but collaborators may accidentally import already loaded
        classes into the datatypes namespace"""
        classes.extend([class_ for class_ in newclasses if class_ not in classes])

    # All kwnown Vis* classes
    _extend(_classes_vis, a99.get_classes_in_module(m, ex.Vis))


def classes_vis():
    """All known Vis* classes"""
    if __flag_first:
        __setup()
    return _classes_vis


_classes_vis = []
__flag_first = True
__collaborators = OrderedDict()  # (("pkg",  pkg),))


def __setup():
    """Will be executed in the first time someone calls classes_*() """
    global __collaborators, __flag_first
    __flag_first = False
    for pkgname in __COLLABORATORS:
        try:
            pkg = importlib.import_module(pkgname)
            a99.get_python_logger().info("Imported collaborator package '{}'".format(pkgname))

            try:
                if hasattr(pkg, "_setup_explorer"):
                    pkg._setup_explorer()
                else:
                    _collect_classes(pkg)

                __collaborators[pkgname] = pkg
            except:
                a99.get_python_logger().exception(
                    "Actually, package '{}' gave error".format(pkgname))
                raise
        except:
            a99.get_python_logger().warning("Failed to import package '{}".format(pkgname))
            pass



import matplotlib.pyplot as plt
from pylab import MaxNLocator
import copy
import numpy as np
from itertools import product, combinations, cycle
import a99
import f311.filetypes as ft


__all__ = ["draw_cube_3d", "draw_cube_colors"]


_ZERO_OFFSET = 0.
def draw_cube_3d(ax, sparsecube, height_threshold=15):
    """
    Plots front and back grid, scaled fluxes, into existing axis

    Args:
        sparsecube: SparseCube instance
        height_threshold: maximum cube height to plot actual spectra.
         If the cube height is greated than this, line segments will be drawn
         instead of plotting the spectra (for speed)

    data cube            mapped to 3D axis
    ------------------   -----------------
    X pixel coordinate   x
    Y pixel coordinate   z
    Z wavelength         y
    """
    assert isinstance(sparsecube, ft.SparseCube)

    flag_segments = sparsecube.height > height_threshold
    flag_empty = len(sparsecube.spectra) == 0
    r0 = [_ZERO_OFFSET, sparsecube.width +_ZERO_OFFSET]
    r2 = [_ZERO_OFFSET, sparsecube.height +_ZERO_OFFSET]
    if flag_empty:
        r1 = [_ZERO_OFFSET, 1+_ZERO_OFFSET]
    else:
        max_flux = max([max(sp.flux) for sp in sparsecube.spectra])
        _y = sparsecube.wavelength
        dlambda = _y[1] - _y[0]
        r1 = [_y[0] - dlambda / 2, _y[-1] + dlambda / 2]
        scale = 1. / max_flux

    PAR = {"color": "y", "alpha": 0.3}

    def draw_line(*args, **kwargs):
        tempdict = copy.copy(PAR)
        tempdict.update(kwargs)
        ax.plot3D(*args, **tempdict)

    # draws cube edges using a thicker line
    if not flag_empty:
        for s, e in combinations(np.array(list(product(r0, r1, r2))), 2):
            if np.sum(s == e) == 2:
                # if np.sum(np.abs(s - e)) == r[1] - r[0]:
                draw_line(*zip(s, e), lw=2)

    # draws grids
    for i in range(sparsecube.width):
        draw_line([i + _ZERO_OFFSET] * 2, [r1[0]] * 2, r2)
        draw_line([i + _ZERO_OFFSET] * 2, [r1[1]] * 2, r2)
    for i in range(sparsecube.height):
        draw_line(r0, [r1[0]] * 2, [i + _ZERO_OFFSET] * 2)
        draw_line(r0, [r1[1]] * 2, [i + _ZERO_OFFSET] * 2)

    for sp in sparsecube.spectra:
        if flag_segments:
            flux1 = sp.flux[(0, -1),] * scale + sp.pixel_y + _ZERO_OFFSET
            ax.plot(np.array([1, 1]) * sp.pixel_x + _ZERO_OFFSET + .5,
                    sp.wavelength[(0, -1),],
                    flux1, color='k')
        else:
            n = len(sp)
            flux1 = sp.flux * scale + sp.pixel_y +_ZERO_OFFSET
            ax.plot(np.ones(n) * sp.pixel_x+_ZERO_OFFSET+.5,
                    sp.wavelength,
                    flux1, color='k')

    # ax.set_aspect("equal")
    ax.set_xlabel("x (pixel)")
    ax.set_ylabel('wavelength ($\AA$)')  # ax.set_ylabel('wavelength ($\AA$)')
    ax.set_zlabel('y (pixel)')

    ax.set_ylim([sparsecube.wavelength[0], sparsecube.wavelength[-1]])
    ax.set_zlim([_ZERO_OFFSET, _ZERO_OFFSET+sparsecube.height])
    ax.set_xlim([_ZERO_OFFSET, _ZERO_OFFSET+sparsecube.width])
    ax.zaxis.set_major_locator(MaxNLocator(integer=True))


def draw_cube_colors(ax, sparsecube, vrange, sqx=None, sqy=None, flag_scale=False, method=0):
    """
    Plots image on existing axis

    Arguments
      ax: matplotlib axis
      sparsecube: SparseCube instance
      vrange: visible range
      sqx: "place spectrum" x
      sqy: "place spectrum" y

    Returns: matplotlib plot object representing square, or None
    """
    from f311 import explorer as ex
    assert isinstance(sparsecube, ft.SparseCube)
    im = ex.sparse_cube_to_colors(sparsecube, vrange, flag_scale, method)
    ax.imshow(im, interpolation="nearest")
    ax.invert_yaxis()
    obj_square = None
    K = .5
    if sqx is not None:
        x0, x1, y0, y1 = sqx - K, sqx + K, sqy - K, sqy + K
        obj_square = ax.plot([x0, x1, x1, x0, x0], [y0, y0, y1, y1, y0],
                             c='w', ls='solid', lw=3, alpha=0.5, zorder=99999)
    ax.set_xlim([-K, sparsecube.width - .5])
    ax.set_ylim([-K, sparsecube.height - .5])
    return obj_square



from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
# from a_WState import WState
# import moldb as db
from .a_WFileMolDB import *
import os
from ..a_XFileMainWindow import *

__all__ = ["XFileMolDB"]


class XFileMolDB(XFileMainWindow):
    def _add_stuff(self):
        XFileMainWindow._add_stuff(self)

        import f311.filetypes as ft
        import f311.explorer as ex

        lv = self.keep_ref(QVBoxLayout(self.gotting))
        me = self.w_moldb = WFileMolDB(self)
        lv.addWidget(me)
        me.changed.connect(self._on_changed)

        # # Synchronized sequences
        _VVV = ft.FileMolDB.description
        self.pages.append(ex.MyPage(
         text_tab="{} (Alt+&1)".format(_VVV),
         text_saveas="Save %s as..." % _VVV,
         text_load="Load %s" % _VVV,
         cls_save=ft.FileMolDB, clss_load=(ft.FileMolDB,), wild="*.sqlite",
         editor=me, flag_autosave=True
        ))

        self.setWindowTitle("Molecular information database editor")

    def wants_auto(self):
        idx = self.w_source.index
        filename = None
        if idx == 0:
            lines = self.w_hitran.data
            if lines:
                filename = "{}.dat".format(lines["header"]["table_name"])

        if filename is None:
            # Default
            filename = a99.new_filename("mol", "dat")

        self.w_out.value = filename



#!/usr/bin/env python


"""Lists MOSAIC Spectrograph modes"""


import argparse
import rows
import io
import logging
from f311 import aosss
import a99


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
    )
    parser.add_argument('search', nargs='?', default=None, type=str,
     help='Search string (optional) (e.g., "HMM")')

    args = parser.parse_args()

    modes = aosss.mosaic.modes

    search_str = args.search.lower() if args.search else None

    modes_filtered = [mode for mode in modes
                      if not args.search or
                         search_str in mode.name.lower() or
                         search_str in mode.abbreviation.lower()]

    dicts = [mode.to_dict() for mode in modes_filtered]

    table = rows.import_from_dicts(dicts)
    fobj = io.StringIO()
    rows.export_to_txt(table, fobj)
    fobj.seek(0)
    for line in fobj:
        print(line.strip('\n'),)



#!/usr/bin/env python

"""
Lists all programs available
"""

import argparse
import os
from collections import OrderedDict
import glob
import sys
import a99
import f311


def _get_subpackage_names():
    """Figures out the names of the subpackages of f311

    Source: http://stackoverflow.com/questions/832004/python-finding-all-packages-inside-a-package
    """

    dir_ = os.path.dirname(f311.__file__)

    def is_package(d):
        d = os.path.join(dir_, d)
        return os.path.isdir(d) and glob.glob(os.path.join(d, '__init__.py*'))

    ret = list(filter(is_package, os.listdir(dir_)))
    ret.sort()
    return ret


def _get_scripts_path(pkgname):
    """Returns full path to scripts directory, assuming it is like .../f311/<pkgname>/scripts"""
    return os.path.join(os.path.dirname(f311.__file__), pkgname, "scripts")


def _get_programs_dict(pkgname_only, flag_protected):
    """Returns dictionary {(package description): [ExeInfo0, ...], ...}"""

    allinfo = OrderedDict()
    pkgnames = []
    for pkgname in _get_subpackage_names():
        if pkgname_only is not None and pkgname != pkgname_only:
                continue
        allinfo["Package '{}'".format(pkgname)] = a99.get_exe_info(_get_scripts_path(pkgname), flag_protected)
        pkgnames.append(pkgname)

    # This can be called (without much shame) a "gambiarra"
    if "pyfant" in pkgnames:
        import f311.pyfant as pf
        allinfo["PFANT Fortran binaries"] = pf.get_fortrans()

    return allinfo


def _format_programs_dict(allinfo, format):
    ret = []
    # will indent listing only if there is more than 1 package listed
    ind = 1 if len(allinfo) > 1 else 0
    for title, exeinfo in allinfo.items():
        if len(exeinfo) == 0:
            continue

        ret.extend(a99.format_h1(title, format) + [""])
        linesp, module_len = a99.format_exe_info(exeinfo, format, ind)
        ret.extend(linesp)

    return ret



if __name__ == "__main__":
    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=a99.SmartFormatter
    )
    parser.add_argument('format', type=str, help='Print format', nargs="?", default="text",
                        choices=["text", "markdown-list", "markdown-table", "rest-list"])
    parser.add_argument('-p', '--pkgname', type=str, help='List programs from this package only', default="(all)")
    parser.add_argument('-l', '--list-packages', action="store_true", help='Lists all sub-packages of f311')
    parser.add_argument('-_', '--protected', action="store_true", help="Includes protected scripts (starting with '_')_")
    args = parser.parse_args()

    if args.list_packages:
        print("\n".join(a99.format_h1("Sub-packages of f311")))
        print("\n".join(_get_subpackage_names()))
        sys.exit()

    pkgname_only = args.pkgname if args.pkgname is not "(all)" else None



    allinfo = _get_programs_dict(pkgname_only, args.protected)
    strlist = _format_programs_dict(allinfo, args.format)
    print("\n".join(strlist))
    print("http://github.com/trevisanj/f311\n")




"""Widget to edit a FileMain object."""

__all__ = ["WFileMolConsts"]

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
import f311.filetypes as ft
from .a_WMolecularConstants import *


class WFileMolConsts(a99.WEditor):
    """
    FileMolConsts editor widget.

    Args:
      parent=None
    """

    @property
    def fcfs(self):
        if self.w_molconsts is None:
            return None
        return self.w_molconsts.fcfs

    def __init__(self, *args, **kwargs):
        a99.WEditor.__init__(self, *args, **kwargs)

        # Internal flag to prevent taking action when some field is updated programatically
        self.flag_process_changes = False


        # # Central layout
        l = self.layout_editor

        w = self.w_molconsts = WMolecularConstants(self.parent_form)
        w.changed.connect(self._on_w_molconsts_changed)
        l.addWidget(w)


        # ## Adds vertically expanding spacer at bottom
        l.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))



        self.setEnabled(False)  # disabled until load() is called
        self.flag_process_changes = True


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Interface

    def set_moldb(self, fobj):
        self.w_molconsts.set_moldb(fobj)

    def _do_load(self, fobj):
        assert isinstance(fobj, ft.FileMolConsts)
        self._f = fobj
        self.w_molconsts.molconsts = fobj.molconsts
        self._flag_valid = True

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Qt override

    def setFocus(self, reason=None):
        """Sets focus to first field. Note: reason is ignored."""
        self.w_molconsts.setFocus()


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Slots

    def _on_w_molconsts_changed(self):
        if self.flag_process_changes:
            self._flag_valid = self.w_molconsts.flag_valid
            self.changed.emit()




DESCR_MULTI = """
Runs pfant for different abundances for each element, then run nulbad for each
pfant result for different FWHMs.

The configuration is read from a .py file.

The user must specify a list of FWHM values for nulbad convolutions, and
a dictionary containing element symbols and respective list containing n_abdif
differential abundances to be used for each element.

pfant will be run n_abdif times, each time adding to each element in ab the i-th
value in the vector for the corresponding element.

nulbad will run n_abdif*n_fwhms times, where n_fwhms is the number of different
FWHMs specified.

The result will be
- several spectra saved as  "<star name><pfant name or counter>.sp"
- several "spectra list" files saved as "cv_<FWHM>.spl". As the file indicates,
  each ".spl" file will have the names of the spectrum files for a specific FWHM.
  .spl files are subject to input for lineplot.py by E.Cantelli
"""


"""
Data representations.
"""

from .basic import *
from .molconsts import *
from .ft_basic import *
from .ft_pyfant import *
from .ft_aosss import *
from .ft_uroboros import *
from .classcatalog import *
from .util import *


#!/usr/bin/env python
# coding: utf8
"""
Script to put together the file moldb.sqlite

**Note** This script is protected, but is very important to put to put together a lot of information
needed for molecular lines conversion (``convmol.py``)
"""

from f311 import filetypes as ft
import f311.convmol as cm
import a99
from collections import OrderedDict
import sqlite3
import os
import sys
from f311 import convmol as cm
from f311 import pyfant as pf
from f311.convmol import NIST_URL

# FileMolecules
filemol = None
moldb = None



def my_info(s):
    a99.get_python_logger().info("[build-moldb] {}".format(s))


def insert_molecules():
    """Inserts data into the 'molecule' table"""

    # Molecules present in PFANT molecules.dat
    MOLECULES = [("MgH", "Magnesium monohydride"),
                 ("C2", "Dicarbon"),
                 ("CN", "Cyano radical"),
                 ("CH", "Methylidyne"),
                 ("NH", "Imidogen"),
                 ("CO", "Carbon monoxide"),
                 ("OH", "Hydroxyl radical"),
                 ("FeH", "Iron hydride"),
                 ("TiO", "Titanium oxide")]

    conn.executemany("insert into molecule (formula, name) values (?,?)", MOLECULES)
    conn.commit()


def insert_systems():
    """Parses the systems out of the molecules descriptions in the PFANT molecular lines file"""


    # Collects PFANT systems
    # Iterates through the PFANT molecules to retrieve its systems.
    # Information is stored as MolConsts objects in which only the formula and system information are set
    rows = []
    for molecule in filemol:
        row = ft.MolConsts()
        row.populate_parse_str(molecule.description)

        print("FORMULA {}".format(row["formula"]))

        id_molecule = conn.execute("select id from molecule where formula = ?",
                                   (row["formula"],)).fetchone()["id"]

        row["id_molecule"] = id_molecule

        # Inserts data into the 'system' table
        id_system = moldb.insert_system_if_does_not_exist(row, "by _build-moldb.insert_systems()")

        # Inserts data into the 'pfantmol' table
        # TODO "s" is redundand, make sure I am not using it
        conn.execute("insert into pfantmol "
                     "(id_system, description, fe, do, am, bm, ua, ub, te, cro) "
                     "values (?,?,?,?,?,?,?,?,?,?)",
                     (id_system, molecule.description, molecule.fe, molecule.do, molecule.am,
                      molecule.bm, molecule.ua, molecule.ub, molecule.te,
                      molecule.cro))

    conn.commit()


def insert_nist_data():
    """Tries to download data from NIST Web Book online"""

    for row in conn.execute("select id, formula from molecule order by id").fetchall():
        id_molecule, formula = row["id"], row["formula"]
        a99.get_python_logger().info("Molecule '{}'...".format(formula))
        try:
            data, _, _ = cm.get_nist_webbook_constants(formula)

            for state in data:
                # **Note** assumes that the columns in data match the
                # (number of columns in the state table - 2) and their order
                conn.execute("insert into state values (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)",
                             [None, id_molecule] + state + [""])

            conn.commit()

        except:
            a99.get_python_logger().exception("Failed for molecule '{}'".format(formula))
    conn.commit()


def load_list_file(filename):
    """Loads a file containing (vl, v2l, FranckCondonFactor)

    ===begin exert===
    00 00 sj0000.dat .8638228E+00
    00 01 sj0001.dat .8712862E-01
    00 02 sj0002.dat .3144656E-01
    ===end exert===
    """


    fcfs = OrderedDict()
    with open(filename, "r") as h:
        for line in h:
            vl, v2l, _, fcf = line.strip().split(" ")
            vl = int(vl)
            v2l = int(v2l)
            fcf = float(fcf)
            fcfs[(vl, v2l)] = fcf
    return fcfs








# Electronic systems
#
# # TODO calculate FCFs for Tio, FeH, C2, CO
SYSTEMS_MAP = (
("CH", "cha.out", "A", 2, 2, "X", 2, 1, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),  # A2Delta - X2Pi
("CH", "chb.out", "B", 2, 0, "X", 2, 1, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),
("CH", "chc.out", "C", 2, 0, "X", 2, 1, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),
("CN", "cnb.out", "B", 2, 0, "X", 2, 0, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),
("NH", "nha.out", "A", 3, 1, "X", 3, 0, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),  # A3Pi - X3Sigma
("OH", "oha.out", "A", 2, 0, "X", 2, 1, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),  # A2Sigma - X2Pi
("MgH", "mgha.out", "A", 2, 1, "X", 2, 0, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),
("MgH", "mghb.out", "B", 2, 0, "X", 2, 0, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),
)


def insert_franck_condon_factors():

    for formula, filename, from_label, from_mult, from_spdf, to_label, to_mult, to_spdf, notes in SYSTEMS_MAP:

        my_info("FCFs for system ({}, {}, {}, {}, {}, {}, {})".format(formula, from_label, from_mult, from_spdf, to_label, to_mult, to_spdf))

        id_molecule = conn.execute("select id from molecule where formula = ?",
                                   (formula,)).fetchone()["id"]


        molconsts = ft.MolConsts()
        molconsts.update(id_molecule = id_molecule, from_label = from_label,
         from_mult = from_mult, from_spdf = from_spdf, to_label = to_label,
         to_mult = to_mult, to_spdf = to_spdf)

        id_system = moldb.insert_system_if_does_not_exist(molconsts,
         notes="by _build-moldb.insert_franck_condon_factors()")

        if filename is not None:
            # Can handle two different file formats
            try:
                a = ft.FileFCF()
                a.load(filename)
                fcf_dict = a.fcfs
            except:
                fcf_dict = load_list_file(filename)
            for (vl, v2l), fcf in fcf_dict.items():
                conn.execute("insert into fcf (id_system, vl, v2l, value) values (?,?,?,?)",
                             (id_system, vl, v2l, fcf))
    conn.commit()


if __name__ == "__main__":

    filename = ft.FileMolDB.default_filename

    if os.path.isfile(filename):
        if a99.yesno("File '{}' already exists, get rid of it and continue".format(filename), True):
            os.unlink(filename)
        else:
            sys.exit()

    yn = a99.yesno("Will now create file '{}'. Continue".format(filename), True)

    if not yn:
        sys.exit()


    moldb = ft.FileMolDB()
    moldb.filename = filename
    my_info("Creating schema...")
    moldb.create_schema()



    filemol = ft.FileMolecules()
    filemol.load(pf.get_pfant_data_path("common", "molecules.dat"))


    conn = moldb.get_conn()
    assert isinstance(conn, sqlite3.Connection)


    my_info("New filename: '{}'".format(moldb.filename))
    my_info("Inserting molecules...")
    insert_molecules()
    my_info("Inserting systems...")
    insert_systems()
    my_info("Inserting Franck-Condon Factors from Bruno Castilho's work...")
    insert_franck_condon_factors()

    my_info("Inserting data from NIST Chemistry Web Book ({})...".format(NIST_URL))
    insert_nist_data()



'''"Adaptive Optics Systems Simulation Support"'''

# # Imports
from .basic import *
from .util import *
from .report import *
from .mosaic import *
from .util import *

# from . import basic
# from . import util
# from . import report
# from . import mosaic


# # Function to access package-specific config file
def get_config():
    """Returns AAConfigObj object that corresponds to file ~/.f311.aosss.conf"""
    import a99
    return a99.get_config_obj(".f311.aosss.conf")





from .. import FilePyConfig, ConfigDict
import a99
from collections import defaultdict

__all__ = ["FileConfigConvMol"]


@a99.froze_it
class FileConfigConvMol(FilePyConfig):
    """Python source containing 'config_conv = ConfigConv(...)"""

    description = "configuration file for molecular lines conversion GUI"
    default_filename = "configconvmol.py"
    attrs = ["obj"]
    editors = []

    # Name of variable in module
    modulevarname = "ccm"



"""Window to edit a scaling factor"""


__all__ = ["XScaleSpectrum"]


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import numpy as np
import a99
import f311.filetypes as ft
import f311.physics as ph


class XScaleSpectrum(a99.XLogDialog):

    def __init__(self, parent=None, file_main=None, file_abonds=None):
        a99.XLogDialog.__init__(self, parent)

        def keep_ref(obj):
            self._refs.append(obj)
            return obj

        self.setWindowTitle("Scale spectrum")

        self.bandpasses = ph.get_ubv_bandpasses()

        # Internal flag to prevent taking action when some field is updated programatically
        self.flag_process_changes = False
        self.spectrum = None  # Spectrum object
        self.cmag = None
        self.factor_ = None

        # # Central layout
        lantanide = self.centralLayout = QVBoxLayout()
        a99.set_margin(lantanide, 0)
        self.setLayout(lantanide)

        # ## Horizontal splitter occupying main area: (options area) | (plot area)
        sp2 = self.splitter2 = QSplitter(Qt.Horizontal)
        lantanide.addWidget(sp2)

        ###
        wleft = keep_ref(QWidget())
        lwleft = QVBoxLayout(wleft)
        a99.set_margin(lwleft, 3)
        ###
        # lwleft.addWidget(keep_ref(QLabel("<b>Reference band</b>")))
        # ###

        lgrid = keep_ref(QGridLayout())
        lwleft.addLayout(lgrid)
        a99.set_margin(lgrid, 0)
        lgrid.setVerticalSpacing(4)
        lgrid.setHorizontalSpacing(5)

        # field map: [(label widget, edit widget, field name, short description, long description), ...]
        map = self._map0 = []
        signals = []  # for the SignalProxy below
        ###
        x = keep_ref(QLabel())
        y = self.cb_band = QComboBox()
        signals.append(y.currentIndexChanged)
        x.setBuddy(y)
        y.addItems([bp.name for bp in self.bandpasses])
        map.append((x, y, "&Band name", "UBVRI-x system", ""))
        ###
        x = keep_ref(QLabel())
        y = self.cb_system = QComboBox()
        signals.append(y.currentIndexChanged)
        x.setBuddy(y)
        y.addItems(["ab", "vega", "stdflux"])
        map.append((x, y, "Magnitude &system",
                   "<b>'ab'</b>: AB[solute]<br>"
                   "<b>'vega'</b>: uses Vega spectrum as reference;<br>"
                   "<b>'stdflux'</b>: uses standard reference values<br>from literature", ""))
        ###
        x = self.label_zero_point = QLabel()
        y = self.spinBox_zero_point = QDoubleSpinBox()
        y.setSingleStep(.01)
        y.setDecimals(3)
        y.setMinimum(-100)
        y.setMaximum(100)
        signals.append(y.valueChanged)
        x.setBuddy(y)
        map.append((x, y, "&Zero point", "Value to subtract from calculated magnitude", ""))
        ###
        x = self.label_x = QLabel()
        y = self.checkbox_force_band_range = QCheckBox()
        signals.append(y.stateChanged)
        x.setBuddy(y)
        map.append((x, y, "&Force filter range?", "(even when spectrum does not<br>completely overlap filter)", ""))
        ###
        x = self.label_x = QLabel()
        y = self.spinBox_mag = QDoubleSpinBox()
        y.setSingleStep(.1)
        y.setDecimals(3)
        y.setMinimum(-100)
        y.setMaximum(100)
        signals.append(y.valueChanged)
        x.setBuddy(y)
        map.append((x, y, "Desired apparent &magnitude", "", ""))
        ###

        for i, (label, edit, name, short_descr, long_descr) in enumerate(map):
            # label.setStyleSheet("QLabel {text-align: right}")
            assert isinstance(label, QLabel)
            label.setText(a99.enc_name_descr(name, short_descr))
            label.setAlignment(Qt.AlignRight)
            lgrid.addWidget(label, i, 0)
            lgrid.addWidget(edit, i, 1)
            label.setToolTip(long_descr)
            edit.setToolTip(long_descr)

        # ### Text Edit To Show calculated results
        x = self.keep_ref(QLabel("<b>Computations</b>"))
        y = self.textEdit = QTextEdit()
        x.setBuddy(y)
        y.setReadOnly(True)
        y.setStyleSheet("QTextEdit {color: %s}" % a99.COLOR_DESCR)
        lwleft.addWidget(x)
        lwleft.addWidget(y)

        bb = keep_ref(QDialogButtonBox())
        bb.setOrientation(Qt.Horizontal)
        bb.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        lwleft.addWidget(bb)

        bb.rejected.connect(self.reject)
        bb.accepted.connect(self.accept)

        # #### Plot area
        wright = keep_ref(QWidget())
        self.figure0, self.canvas0, self.lfig0 = a99.get_matplotlib_layout(wright)

        sp2.addWidget(wleft)
        sp2.addWidget(wright)

        # # Signal proxy
        # Limits the number of times that on_sth_changed() is called
        self.signal_proxy = a99.SignalProxy(signals, delay=0.3, rateLimit=0, slot=self.on_sth_changed)

        self.setEnabled(False)  # disabled until load() is called
        a99.style_checkboxes(self)
        self.flag_process_changes = True

        # self.__update()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Getters & setters


    def band_index(self):
        return self.cb_band.currentIndex()

    def band_name(self):
        return str(self.cb_band.currentText())

    def set_spectrum(self, x):
        assert isinstance(x, ft.Spectrum)
        self.spectrum = x
        self.setEnabled(True)
        self.__update()

    def set_band_name(self, x):
        self.cb_band.setEditText(x)
        self.__update()

    def system(self):
        return str(self.cb_system.currentText())

    def set_system(self, x):
        self.cb_system.setEditText(x)
        self.__update()

    def flag_force_band_range(self):
        return self.checkbox_force_band_range.isChecked()

    def set_flag_force_band_range(self, x):
        self.checkbox_force_band_range.setChecked(bool(x))
        self.__update()

    def desired_magnitude(self):
        return self.spinBox_mag.value()

    def zero_point(self):
        return self.spinBox_zero_point.value()

    def calculated_magnitude(self):
        return self.cmag

    def factor(self):
        return self.factor_

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Qt override

    def accept(self):
        if self.factor_ is None or np.isinf(self.factor_):
            _ = QMessageBox.critical(None, "Cannot scale", "Scaling factor cannot be calculated")
            return False
        QDialog.accept(self)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Slots

    def on_sth_changed(self):
        if self.flag_process_changes:
            self.__update()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Internal gear
            

    def __update(self):
        self.flag_process_changes = False
        try:
            fig = self.figure0
            fig.clear()
            # name = self.band_name()
            bandpass = self.bandpasses[self.band_index()]
            mag_data = ph.calculate_magnitude(self.spectrum, bandpass, self.system(), self.zero_point(),
                                              self.flag_force_band_range())
            _draw_figure(fig, mag_data, self.spectrum, self.flag_force_band_range())
            self.canvas0.draw()

            mag = self.desired_magnitude()
            cmag = mag_data["cmag"]

            self.factor_ = k = ph.MAGNITUDE_BASE ** (cmag - mag) if cmag is not None else float("nan")

            # Updates calculated state

            kk = mag_data.keys()
            lenk = max([len(k) for k in kk])
            vv = mag_data.values()
            text = "\n".join(["{0:{1}}: {2}".format(k, lenk, v) for k, v in mag_data.items()])


            self.textEdit.setText("<pre>\n%s\n\n"
                                  "Scaling factor: %g"
                                  "</pre>" % (text, self.factor_))

        except Exception as E:
            self.add_log_error(str(E))
            a99.get_python_logger().exception("Could not plot band_curve")
        finally:
            self.flag_process_changes = True



def _draw_figure(fig, mag_data, spectrum, flag_force_band_range):
    from f311 import explorer as ex

    # y = s*f ; s and y: fluxes ; f: filter ; all functions of wavelength
    # out_area = integrate y over whole axis, but y = 0 outside the filter range
    # weighted_mean_flux = out_area/band_area


    bp = mag_data["bp"]
    weighted_mean_flux = mag_data["weighted_mean_flux"]
    calc_l0, calc_lf = mag_data["calc_l0"], mag_data["calc_lf"]
    filtered_sp = mag_data["filtered_sp"]
    band_area = bp.area(calc_l0, calc_lf)
    out_y = filtered_sp.y
    out_area = np.trapz(filtered_sp.y, filtered_sp.x)
    cmag = mag_data["cmag"]

    MARGIN_H = .15  # extends horizontally beyond band range
    MARGIN_V = .2  # extends vertically beyond band range
    COLOR_OTHER_BANDS = (.4, .4, .4)
    COLOR_CURRENT_BAND = (.1, .1, .1)
    COLOR_FILL_BAND = (.9, .8, .8)
    LINE_WIDTH = 1.5
    band_x = np.linspace(bp.l0, bp.lf, 200)
    band_y = bp.ufunc()(band_x)
    band_span_x = bp.lf - bp.l0
    band_max_y = max(band_y)
    plot_l0, plot_lf = bp.l0 - band_span_x * MARGIN_H, bp.lf + band_span_x * MARGIN_H
    plot_h_middle = (plot_l0 + plot_lf) / 2
    spp = ex.SB_Cut(plot_l0, plot_lf).use(spectrum)  # spectrum for plotting
    flux_ylim = [0, np.max(spp.y) * (1 + MARGIN_V)] if len(spp) > 0 else [-.1, .1]

    # # First subplot
    ax = ax0 = fig.add_subplot(311)
    if len(spp) == 0:
        ax.plot([], [])
        ax.annotate("Band out of spectral range [%g, %g]" % (spectrum.x[0], spectrum.x[-1]), xy=(plot_h_middle, 0),
                    horizontalalignment="center", verticalalignment="center")
    else:
        ax.plot(spp.x, spp.y, c=COLOR_CURRENT_BAND, lw=LINE_WIDTH)
        # shows weighted mean flux within range (flat spectrum that would have the same
        # magnitude as the actual spectrum)
        ax.plot([calc_l0, calc_lf], [weighted_mean_flux, weighted_mean_flux], linestyle="dashed",
                linewidth=LINE_WIDTH, color=(.4, .3, .1))
    # ax.plot(spc.x, spc.y, c=COLOR_CURRENT_BAND, lw=LINE_WIDTH, zorder=999)
    ax.set_ylim(flux_ylim)
    ax.set_ylabel("Flux")

    # # Second subplot
    overall_max_y = 0
    ax = fig.add_subplot(312, sharex=ax0)
    # other bands
    for band in ph.get_ubv_bandpasses():
        if band.lf >= plot_l0 and band.l0 <= plot_lf:
            x = np.linspace(band.l0, band.lf, 200)
            y = band.ufunc()(x)
            ax.plot(x, y, lw=LINE_WIDTH)  # c=COLOR_OTHER_BANDS)
            idx_max = np.argmax(y)
            max_y = y[idx_max]
            overall_max_y = max(max_y, band_max_y)
            ax.annotate(band.name, xy=(x[idx_max], y[idx_max] + .1),
                        horizontalalignment="center", verticalalignment="center")
    # current band
    ax.plot(band_x, band_y, c=COLOR_CURRENT_BAND, lw=LINE_WIDTH)
    xfill = np.linspace(calc_l0, calc_lf, 50)
    yfill = bp.ufunc()(xfill)
    ax.fill_between(xfill, yfill, color=COLOR_FILL_BAND)
    ax.set_ylim([0, overall_max_y * (1 + MARGIN_V)])
    ax.set_ylabel("Bandpass filter curve")
    idx_max = np.argmax(band_y)
    ax.annotate("A1=%.1f" % band_area, xy=(band_x[idx_max], band_max_y * .45),
                horizontalalignment="center", verticalalignment="center",
                color=(.4, .2, .1))

    # Third subplot
    # # First subplot
    ax = fig.add_subplot(313, sharex=ax0)
    if len(filtered_sp) == 0:
        ax.plot([], [])
        ax.annotate("Band out of spectral range [%g, %g]" % (spectrum.x[0], spectrum.x[-1]), xy=(plot_h_middle, 0),
                    horizontalalignment="center", verticalalignment="center")
    else:
        ax.plot(filtered_sp.x, out_y, c=COLOR_CURRENT_BAND, lw=LINE_WIDTH)
        ax.fill_between(filtered_sp.x, out_y, color=COLOR_FILL_BAND)
        
        ax.plot(xfill, yfill*weighted_mean_flux, linestyle="dashed",
                linewidth=LINE_WIDTH, color=(.4, .3, .1), label='equivalent flat')
        ax.annotate("A2=%g" % out_area, xy=((xfill[0] + xfill[-1]) / 2, max(out_y) * .45),
                    horizontalalignment="center", verticalalignment="center",
                    color=(.4, .2, .1))
    ax.set_xlim([plot_l0, plot_lf])
    ax.set_ylabel("Filtered flux")
    ax.set_ylim(flux_ylim)
    ax.set_xlabel("Wavelenght ($\AA$)")

    fig.tight_layout()

    return weighted_mean_flux, cmag



def _toggle_widget(w, flag_readonly):
    w.setReadOnly(flag_readonly)
    sheet = ""
    if flag_readonly:
        sheet = "QWidget {background: #A09D9D}"
    w.setStyleSheet(sheet)




import a99
from .. import DataFile
from astropy.io import fits
from collections import defaultdict, Counter

__all__ = ["FileGalfit"]

@a99.froze_it
class FileGalfit(DataFile):
    """FITS file with frames named INPUT_*, MODEL_*, RESIDUAL_*, which is the output of Galfit software

    When file is loaded, the band_names property will be filled

    IF INPUT_*, MODEL_* and RESIDUAL_* do not have the same band names, the file is "rejected"
    """
    attrs = []
    description = "Galfit output"
    default_filename = None
    flag_txt = False

    @property
    def band_names(self):
        return self._band_names

    @property
    def kind_names(self):
        return self._kind_names

    def __init__(self):
        DataFile.__init__(self)
        self.hdulist = None
        self._band_names = []
        self._kind_names = []
        # Frame indexes by key (kind_name, band_name)
        self._idxs = {}

    def get_frame(self, kind_name, band_name):
        """Convenience function to get frame using its "kind name", "band name"

        Args:
            kind_name: str in self.kind_names
            band_name: str in self.band_names
        """
        return self.hdulist[self._idxs[(kind_name, band_name)]]



    def _do_load(self, filename):
        f = fits.open(filename)

        info = f.info(False)


        # Example:
        #     [(0, 'PRIMARY', 'PrimaryHDU', 8, (720, 720), 'float32', ''),
        #      (1, 'INPUT_fuv', 'ImageHDU', 40, (720, 720), 'float64', ''),
        #      (2, 'INPUT_nuv', 'ImageHDU', 40, (720, 720), 'float64', ''),
        #      (3, 'INPUT_r', 'ImageHDU', 42, (720, 720), 'float64', ''),
        #      (4, 'INPUT_j', 'ImageHDU', 41, (720, 720), 'float64', ''),
        #      (5, 'INPUT_h', 'ImageHDU', 41, (720, 720), 'float64', ''),
        #      (6, 'INPUT_k', 'ImageHDU', 41, (720, 720), 'float64', ''),
        #      (7, 'INPUT_3.4', 'ImageHDU', 35, (720, 720), 'float64', ''),
        #      (8, 'INPUT_4.6', 'ImageHDU', 35, (720, 720), 'float64', ''),
        #      (9, 'MODEL_fuv', 'ImageHDU', 409, (720, 720), 'float32', ''),
        #      (10, 'MODEL_nuv', 'ImageHDU', 409, (720, 720), 'float32', '')

        kinds = defaultdict(lambda: [])
        kind_names = ("INPUT", "MODEL", "RESIDUAL")
        counts = Counter()
        idxs = {}

        for record in info[1:]:  # skips "PRIMARY"
            index = record[0]
            name = record[1]
            if name.startswith(kind_names):
                counts[kind_name] += 1
                kind_name, band_name = name.split("_")
                kinds[kind_name].append(band_name)
                idxs[(kind_name, band_name)] = index

        # Validation: must find at least one INPUT, one MODEL, and one RESIDUAL
        for kind_name in kind_names:
            if kind_name not in counts:
                raise RuntimeError("Cannot find a frame containing the word '{}' in its name".format(kind_name))

        # Validation: band names of all kinds must match
        if kinds["INPUT"] != kinds["MODEL"]:
            raise RuntimeError("(INPUT, MODEL) band names mismatch: {} != {}".
                               format(kinds["INPUT"], kinds["MODEL"]))
        if kinds["INPUT"] != kinds["RESIDUAL"]:
            raise RuntimeError("(INPUT, RESIDUAL) band names mismatch: {} != {}".
                               format(kinds["INPUT"], kinds["RESIDUAL"]))


        self._band_names = kinds["INPUT"]
        self._kind_names = kind_names
        self._idxs = idxs

        self.hdulist = f

    def _do_save_as(self, filename):
        a99.overwrite_fits(self.hdulist, filename)



from PyQt5.QtWidgets import *
from .a_XFileMainWindowBase import *

__all__ = ["XFileMainWindow", "NullEditor"]

class XFileMainWindow(XFileMainWindowBase):
    """Application template with file operations in two tabs: (application area) and (log)"""

    def __init__(self, *args, **kwargs):
        XFileMainWindowBase.__init__(self, *args, **kwargs)

        self._add_log_tab()
        # TODO make this flexible if needed
        self.tabWidget.setCurrentIndex(0)

    def _add_stuff(self):
        # ### Custom Area tab will have a widget **without** a layout, named "gotting"
        w0 = self.gotting = QWidget()
        self.tabWidget.addTab(w0, "")
